import os
import time

from . import get_keras_version
from .layers.batchnorm import MakeKerasBatchNorm
from .layers.binary import MakeKerasBinary
from .layers.concat import MakeKerasConcat
from .layers.conv import MakeKerasConv
from .layers.dense import MakeKerasDense
from .layers.elu import MakeKerasELU
from .layers.flatten import MakeKerasFlatten
from .layers.layernorm import MakeKerasLayerNorm
from .layers.leaky_relu import MakeKerasLeakyRelu
from .layers.permute import MakeKerasPermute
from .layers.pooling import MakeKerasPooling
from .layers.relu import MakeKerasReLU
from .layers.reshape import MakeKerasReshape
from .layers.selu import MakeKerasSeLU
from .layers.sigmoid import MakeKerasSigmoid
from .layers.softmax import MakeKerasSoftmax
from .layers.swish import MakeKerasSwish
from .layers.tanh import MakeKerasTanh


def MakeKerasActivation(layer):
    attributes = layer["layerAttributes"]
    activation = attributes["activation"]
    fLayerActivation = str(activation.__name__)

    if fLayerActivation in mapKerasLayer.keys():
        return mapKerasLayer[fLayerActivation](layer)
    else:
        raise Exception("TMVA.SOFIE - parsing keras activation layer " + fLayerActivation + " is not yet supported")


# Set global dictionaries, mapping layers to corresponding functions that create their ROperator instances
mapKerasLayer = {
    "Activation": MakeKerasActivation,
    "Permute": MakeKerasPermute,
    "BatchNormalization": MakeKerasBatchNorm,
    "LayerNormalization": MakeKerasLayerNorm,
    "Reshape": MakeKerasReshape,
    "Flatten": MakeKerasFlatten,
    "Concatenate": MakeKerasConcat,
    "swish": MakeKerasSwish,
    "silu": MakeKerasSwish,
    "Add": MakeKerasBinary,
    "Subtract": MakeKerasBinary,
    "Multiply": MakeKerasBinary,
    "Softmax": MakeKerasSoftmax,
    "tanh": MakeKerasTanh,
    #  "Identity": MakeKerasIdentity,
    #  "Dropout": MakeKerasIdentity,
    "ReLU": MakeKerasReLU,
    "relu": MakeKerasReLU,
    "ELU": MakeKerasELU,
    "elu": MakeKerasELU,
    "selu": MakeKerasSeLU,
    "sigmoid": MakeKerasSigmoid,
    "LeakyReLU": MakeKerasLeakyRelu,
    "leaky_relu": MakeKerasLeakyRelu,
    "softmax": MakeKerasSoftmax,
    "MaxPooling2D": MakeKerasPooling,
    "AveragePooling2D": MakeKerasPooling,
    "GlobalAveragePooling2D": MakeKerasPooling,
    #  "SimpleRNN": MakeKerasRNN,
    #  "GRU": MakeKerasRNN,
    #  "LSTM": MakeKerasRNN,
}

mapKerasLayerWithActivation = {"Dense": MakeKerasDense, "Conv2D": MakeKerasConv}


def add_layer_into_RModel(rmodel, layer_data):
    """
    Add a Keras layer operation to an existing RModel using the SOFIE framework.

    This function takes an existing RModel and a dictionary representing a Keras layer
    and its attributes, and adds the corresponding layer operation to the RModel using
    the SOFIE framework. The function supports various types of Keras layers, including
    those with or without activation functions.

    Parameters:
    rmodel (RModel): An existing RModel to which the layer operation will be added.
    layer_data (dict): A dictionary containing layer information including type,
                      attributes, input, output, and layer data type.

    Returns:
    RModel: The updated RModel after adding the layer operation.

    Raises exception: If the provided layer type or activation function is not supported.
    """
    import numpy as np
    from ROOT.TMVA.Experimental import SOFIE

    def move_operator(op):
        """
        Wrap an operator into a std::unique_ptr to pass it to RModel::AddOperator().
        """
        import ROOT

        ROOT.SetOwnership(op, False)
        return ROOT.std.unique_ptr[type(op)](op)

    keras_version = get_keras_version()

    fLayerType = layer_data["layerType"]

    # reshape and flatten layers don't have weights, but they need constant tensor for the shape
    if fLayerType == "Reshape" or fLayerType == "Flatten":
        Attributes = layer_data["layerAttributes"]
        if keras_version < "2.16":
            LayerName = Attributes["_name"]
        else:
            LayerName = Attributes["name"]

        if fLayerType == "Reshape":
            TargetShape = np.asarray(Attributes["target_shape"]).astype("int64")
            TargetShape = np.insert(TargetShape, 0, 1)
        else:
            if "_build_input_shape" in Attributes.keys():
                input_shape = Attributes["_build_input_shape"]
            elif "_build_shapes_dict" in Attributes.keys():
                input_shape = list(Attributes["_build_shapes_dict"]["input_shape"])
            else:
                raise RuntimeError("Failed to extract build input shape from " + fLayerType + " layer")
            TargetShape = [SOFIE.ConvertShapeToLength(input_shape[1:])]
            TargetShape = np.asarray(TargetShape)

        # since the AddInitializedTensor method in RModel requires unique pointer, we call a helper function
        # in c++ that does the conversion from a regular pointer to unique one in c++
        # print('adding initialized tensor..',LayerName, TargetShape)
        shape_tensor_name = LayerName + "_shape"
        shape_data = TargetShape.data
        print(TargetShape, shape_data)
        print(len(TargetShape))
        rmodel.AddInitializedTensor["int64_t"](shape_tensor_name, [len(TargetShape)], shape_data)

    # These layers only have one operator - excluding the recurrent layers, in which the activation function(s)
    # are included in the recurrent operator
    if fLayerType in mapKerasLayer.keys():
        Attributes = layer_data["layerAttributes"]
        inputs = layer_data["layerInput"]
        outputs = layer_data["layerOutput"]
        if keras_version < "2.16":
            LayerName = Attributes["_name"]
        else:
            LayerName = Attributes["name"]

        # Convoltion/Pooling layers in keras by default assume the channels dimension is the
        # last one, while in onnx (and the SOFIE's RModel) it is the first one (other than batch
        # size), so a transpose is needed before and after the pooling, if the data format is
        # channels last (can be set to channels first by the user). In case of MaxPool2D and
        # Conv2D (with linear activation) channels last, the transpose layers are added as:

        #                   input                   output
        # transpose layer   input_layer_name        layer_name + PreTrans
        # actual layer      layer_name + PreTrans   layer_name + PostTrans
        # transpose layer   layer_name + PostTrans  output_layer_name

        fLayerOutput = outputs[0]
        if fLayerType == "GlobalAveragePooling2D":
            if layer_data["channels_last"]:
                op = SOFIE.ROperator_Transpose([0, 3, 1, 2], inputs[0], LayerName + "PreTrans")
                rmodel.AddOperator(move_operator(op))
                inputs[0] = LayerName + "PreTrans"
            outputs[0] = LayerName + "Squeeze"
            rmodel.AddOperator(move_operator(mapKerasLayer[fLayerType](layer_data)))
            op = SOFIE.ROperator_Reshape(SOFIE.ReshapeOpMode.Squeeze, [2, 3], LayerName + "Squeeze", fLayerOutput)
            rmodel.AddOperator(move_operator(op))

        # Similar case is with Batchnorm, ONNX assumes that the 'axis' is always 1, but Keras
        # gives the user the choice of specifying it. So, we have to transpose the input layer
        # as 'axis' as the first dimension, apply the BatchNormalization operator and then
        # again tranpose it to bring back the original dimensions
        elif fLayerType == "BatchNormalization":
            if "_build_input_shape" in Attributes.keys():
                num_input_shapes = len(Attributes["_build_input_shape"])
            elif "_build_shapes_dict" in Attributes.keys():
                num_input_shapes = len(list(Attributes["_build_shapes_dict"]["input_shape"]))

            axis = Attributes["axis"]
            axis = axis[0] if isinstance(axis, list) else axis
            if axis < 0:
                axis += num_input_shapes
            fAttrPerm = list(range(0, num_input_shapes))
            fAttrPerm[1] = axis
            fAttrPerm[axis] = 1
            op = SOFIE.ROperator_Transpose(fAttrPerm, inputs[0], LayerName + "PreTrans")
            rmodel.AddOperator(move_operator(op))
            inputs[0] = LayerName + "PreTrans"
            outputs[0] = LayerName + "PostTrans"
            rmodel.AddOperator(move_operator(mapKerasLayer[fLayerType](layer_data)))
            op = SOFIE.ROperator_Transpose(fAttrPerm, LayerName + "PostTrans", fLayerOutput)
            rmodel.AddOperator(move_operator(op))

        elif fLayerType == "MaxPooling2D" or fLayerType == "AveragePooling2D":
            if layer_data["channels_last"]:
                op = SOFIE.ROperator_Transpose([0, 3, 1, 2], inputs[0], LayerName + "PreTrans")
                rmodel.AddOperator(move_operator(op))
                inputs[0] = LayerName + "PreTrans"
                outputs[0] = LayerName + "PostTrans"
            rmodel.AddOperator(move_operator(mapKerasLayer[fLayerType](layer_data)))
            if layer_data["channels_last"]:
                op = SOFIE.ROperator_Transpose([0, 2, 3, 1], LayerName + "PostTrans", fLayerOutput)
                rmodel.AddOperator(move_operator(op))

        else:
            rmodel.AddOperator(move_operator(mapKerasLayer[fLayerType](layer_data)))

        return rmodel

    # These layers require two operators - dense/conv and their activation function
    elif fLayerType in mapKerasLayerWithActivation.keys():
        Attributes = layer_data["layerAttributes"]
        if keras_version < "2.16":
            LayerName = Attributes["_name"]
        else:
            LayerName = Attributes["name"]
        fPActivation = Attributes["activation"]
        LayerActivation = fPActivation.__name__
        if LayerActivation in ["selu", "sigmoid"]:
            rmodel.AddNeededStdLib("cmath")

        # if there is an activation function after the layer
        if LayerActivation != "linear":
            if LayerActivation not in mapKerasLayer.keys():
                raise Exception(
                    "TMVA.SOFIE - parsing keras activation function " + LayerActivation + " is not yet supported"
                )
            outputs = layer_data["layerOutput"]
            inputs = layer_data["layerInput"]
            fActivationLayerOutput = outputs[0]

            # like pooling, convolutional layer from keras requires transpose before and after to match
            # the onnx format
            # if the data format is channels last (can be set to channels first by the user).
            if fLayerType == "Conv2D":
                if layer_data["channels_last"]:
                    op = SOFIE.ROperator_Transpose([0, 3, 1, 2], inputs[0], LayerName + "PreTrans")
                    rmodel.AddOperator(move_operator(op))
                    inputs[0] = LayerName + "PreTrans"
                    layer_data["layerInput"] = inputs
            outputs[0] = LayerName + fLayerType
            layer_data["layerOutput"] = outputs
            op = mapKerasLayerWithActivation[fLayerType](layer_data)
            rmodel.AddOperator(move_operator(op))
            Activation_layer_input = LayerName + fLayerType
            if fLayerType == "Conv2D":
                if layer_data["channels_last"]:
                    op = SOFIE.ROperator_Transpose(
                        [0, 2, 3, 1], LayerName + fLayerType, LayerName + "PostTrans"
                    )
                    rmodel.AddOperator(move_operator(op))
                    Activation_layer_input = LayerName + "PostTrans"

            # Adding the activation function
            inputs[0] = Activation_layer_input
            outputs[0] = fActivationLayerOutput
            layer_data["layerInput"] = inputs
            layer_data["layerOutput"] = outputs

            rmodel.AddOperator(move_operator(mapKerasLayer[LayerActivation](layer_data)))

        else:  # if layer is conv and the activation is linear, we need to add transpose before and after
            if fLayerType == "Conv2D":
                inputs = layer_data["layerInput"]
                outputs = layer_data["layerOutput"]
                fLayerOutput = outputs[0]
                if layer_data["channels_last"]:
                    op = SOFIE.ROperator_Transpose([0, 3, 1, 2], inputs[0], LayerName + "PreTrans")
                    rmodel.AddOperator(move_operator(op))
                    inputs[0] = LayerName + "PreTrans"
                    layer_data["layerInput"] = inputs
                    outputs[0] = LayerName + "PostTrans"
            rmodel.AddOperator(move_operator(mapKerasLayerWithActivation[fLayerType](layer_data)))
            if fLayerType == "Conv2D":
                if layer_data["channels_last"]:
                    op = SOFIE.ROperator_Transpose([0, 2, 3, 1], LayerName + "PostTrans", fLayerOutput)
                    rmodel.AddOperator(move_operator(op))
        return rmodel
    else:
        raise Exception("TMVA.SOFIE - parsing keras layer " + fLayerType + " is not yet supported")


class PyKeras:
    def Parse(filename, batch_size=1):  # If a model does not have a defined batch size, then assuming it is 1

        # TensorFlow/Keras is too fragile to import unconditionally. As its presence might break several ROOT
        # usecases and importing keras globally will slow down importing ROOT, which is not desired. For this,
        # we import keras within the functions instead of importing it at the start of the file (i.e. globally).
        # So, whenever the parser function is called, only then keras will be imported, and not everytime we
        # import ROOT. Also, we can import keras in multiple functions as many times as we want since Python
        # caches the imported packages.

        import keras
        import numpy as np
        from ROOT.TMVA.Experimental import SOFIE

        keras_version = get_keras_version()

        # Check if file exists
        if not os.path.exists(filename):
            raise RuntimeError("Model file {} not found!".format(filename))

        # load model
        keras_model = keras.models.load_model(filename)
        keras_model.load_weights(filename)

        # create new RModel object
        sep = "/"
        if os.name == "nt":
            sep = "\\"

        isep = filename.rfind(sep)
        filename_nodir = filename
        if isep != -1:
            filename_nodir = filename[isep + 1 :]

        ttime = time.time()
        gmt_time = time.gmtime(ttime)
        parsetime = time.asctime(gmt_time)

        rmodel = SOFIE.RModel(filename_nodir, parsetime)

        print("PyKeras: parsing model ", filename)

        # iterate over the layers and add them to the RModel
        # in case of keras 3.x (particularly in sequential models), the layer input and output name conventions are
        # different from keras 2.x. In keras 2.x, the layer input name is consistent with previous layer's output
        # name. For e.g., if the sequence of layers is dense -> maxpool, the input and output layer names would be:
        #           layer   |       name
        # input     dense   |   keras_tensor_1
        # output    dense   |   keras_tensor_2 --
        #                   |                    |=> layer names match
        # input     maxpool |   keras_tensor_2 --
        # output    maxpool |   keras_tensor_3
        #
        # but in case of keras 3.x, this changes.
        #           layer   |       name
        # input     dense   |   keras_tensor_1
        # output    dense   |   keras_tensor_2 --
        #                   |                    |=> different layer names
        # input     maxpool |   keras_tensor_3 --
        # output    maxpool |   keras_tensor_4
        #
        # hence, we need to add a custom layer iterator, which would replace the suffix of the layer's input
        # and output names
        layer_iter = 0
        is_functional_model = True if keras_model.__class__.__name__ == "Functional" else False
        for layer in keras_model.layers:
            layer_data = {}
            layer_data["layerType"] = layer.__class__.__name__
            layer_data["layerAttributes"] = layer.__dict__
            # get input names for layer
            if keras_version < "2.16" or is_functional_model:
                if "input_layer" in layer.name:
                    layer_data["layerInput"] = layer.name
                else:
                    layer_data["layerInput"] = (
                        [x.name for x in layer.input] if isinstance(layer.input, list) else [layer.input.name]
                    )
            else:
                # case of Keras3 Sequential model : in this case output of layer is input to following one, but names can be different
                if "input_layer" in layer.input.name:
                    layer_data["layerInput"] = [layer.input.name]
                else:
                    if layer_iter == 0:
                        input_layer_name = "tensor_input_" + layer.name
                    else:
                        input_layer_name = "tensor_output_" + keras_model.layers[layer_iter - 1].name
                    layer_data["layerInput"] = [input_layer_name]
            # get output names of layer
            if keras_version < "2.16" or is_functional_model:
                layer_data["layerOutput"] = (
                    [x.name for x in layer.output] if isinstance(layer.output, list) else [layer.output.name]
                )
            else:
                # sequential model in Keras3
                output_layer_name = "tensor_output_" + layer.name
                layer_data["layerOutput"] = (
                    [x.name for x in layer.output] if isinstance(layer.output, list) else [output_layer_name]
                )

            layer_iter += 1
            fLayerType = layer_data["layerType"]
            layer_data["layerDType"] = layer.dtype

            if len(layer.weights) > 0:
                if keras_version < "2.16":
                    layer_data["layerWeight"] = [x.name for x in layer.weights]
                else:
                    layer_data["layerWeight"] = [x.path for x in layer.weights]
            else:
                layer_data["layerWeight"] = []

            # for convolutional and pooling layers we need to know the format of the data
            if layer_data["layerType"] in ["Conv2D", "MaxPooling2D", "AveragePooling2D", "GlobalAveragePooling2D"]:
                layer_data["channels_last"] = True if layer.data_format == "channels_last" else False

            # for recurrent type layers we need to extract additional unique information
            if layer_data["layerType"] in ["SimpleRNN", "LSTM", "GRU"]:
                layer_data["layerAttributes"]["activation"] = layer.activation
                layer_data["layerAttributes"]["direction"] = "backward" if layer.go_backwards else "forward"
                layer_data["layerAttributes"]["units"] = layer.units
                layer_data["layerAttributes"]["layout"] = layer.input.shape[0] is None
                layer_data["layerAttributes"]["hidden_size"] = layer.output.shape[-1]

                # for GRU and LSTM we need to extract an additional activation function
                if layer_data["layerType"] != "SimpleRNN":
                    layer_data["layerAttributes"]["recurrent_activation"] = layer.recurrent_activation

                # for GRU there are two variants of the reset gate location, we need to know which one is it
                if layer_data["layerType"] == "GRU":
                    layer_data["layerAttributes"]["linear_before_reset"] = (
                        1 if layer.reset_after and layer.recurrent_activation.__name__ == "sigmoid" else 0
                    )

            # Ignoring the input layer of the model
            if fLayerType == "InputLayer":
                continue

            # Adding any required routines depending on the Layer types for generating inference code.
            if fLayerType == "Dense":
                rmodel.AddBlasRoutines({"Gemm", "Gemv"})
            elif fLayerType == "BatchNormalization":
                rmodel.AddBlasRoutines({"Copy", "Axpy"})
            elif fLayerType == "Conv1D" or fLayerType == "Conv2D" or fLayerType == "Conv3D":
                rmodel.AddBlasRoutines({"Gemm", "Axpy"})
            rmodel = add_layer_into_RModel(rmodel, layer_data)

        # Extracting model's weights
        weight = []
        for idx in range(len(keras_model.get_weights())):
            weightProp = {}
            if keras_version < "2.16":
                weightProp["name"] = keras_model.weights[idx].name
            else:
                weightProp["name"] = keras_model.weights[idx].path
            weightProp["dtype"] = keras_model.get_weights()[idx].dtype.name
            if "conv" in weightProp["name"] and keras_model.weights[idx].shape.ndims == 4:
                weightProp["value"] = keras_model.get_weights()[idx].transpose((3, 2, 0, 1)).copy()
            else:
                weightProp["value"] = keras_model.get_weights()[idx]
            weight.append(weightProp)

        # Traversing through all the Weight tensors
        for weightIter in range(len(weight)):
            fWeightTensor = weight[weightIter]
            fWeightName = fWeightTensor["name"]
            fWeightDType = SOFIE.ConvertStringToType(fWeightTensor["dtype"])
            fWeightTensorValue = fWeightTensor["value"]
            fWeightTensorSize = 1
            fWeightTensorShape = []

            # IS IT BATCH SIZE? CHECK ONNX
            if (
                "simple_rnn" in fWeightName
                or "lstm" in fWeightName
                or ("gru" in fWeightName and "bias" not in fWeightName)
            ):
                fWeightTensorShape.append(1)

            # Building the shape vector and finding the tensor size
            for j in range(len(fWeightTensorValue.shape)):
                fWeightTensorShape.append(fWeightTensorValue.shape[j])
                fWeightTensorSize *= fWeightTensorValue.shape[j]

            if fWeightDType == SOFIE.ETensorType.FLOAT:
                fWeightArray = fWeightTensorValue

                # weights conversion format between keras and onnx for lstm: the order of the different
                # elements (input, output, forget, cell) inside the vector/matrix is different
                if "lstm" in fWeightName:
                    if "kernel" in fWeightName:
                        units = int(fWeightArray.shape[1] / 4)
                        W_f = fWeightArray[:, units : units * 2].copy()
                        W_c = fWeightArray[:, units * 2 : units * 3].copy()
                        W_o = fWeightArray[:, units * 3 :].copy()
                        fWeightArray[:, units : units * 2] = W_o
                        fWeightArray[:, units * 2 : units * 3] = W_f
                        fWeightArray[:, units * 3 :] = W_c
                    else:  # bias
                        units = int(fWeightArray.shape[0] / 4)
                        W_f = fWeightArray[units : units * 2].copy()
                        W_c = fWeightArray[units * 2 : units * 3].copy()
                        W_o = fWeightArray[units * 3 :].copy()
                        fWeightArray[units : units * 2] = W_o
                        fWeightArray[units * 2 : units * 3] = W_f
                        fWeightArray[units * 3 :] = W_c

                # need to make specific adjustments for recurrent weights and biases
                if "simple_rnn" in fWeightName or "lstm" in fWeightName or "gru" in fWeightName:
                    # reshaping weight matrices for recurrent layers due to keras-onnx inconsistencies
                    if "kernel" in fWeightName:
                        fWeightArray = np.transpose(fWeightArray)
                        fWeightTensorShape[1], fWeightTensorShape[2] = fWeightTensorShape[2], fWeightTensorShape[1]

                    fData = fWeightArray.flatten()

                    # the recurrent bias and the cell bias can be the same, in which case we need to add a
                    # vector of zeros for the recurrent bias
                    if "bias" in fWeightName and len(fData.shape) == 1:
                        fWeightTensorShape[1] *= 2
                        fRbias = fData.copy() * 0
                        fData = np.concatenate((fData, fRbias))

                else:
                    fData = fWeightArray.flatten()
                rmodel.AddInitializedTensor["float"](fWeightName, fWeightTensorShape, fData)
            else:
                raise TypeError("Type error: TMVA SOFIE does not yet support data layer type: " + fWeightDType)

        # Extracting input tensor info
        if keras_version < "2.16":
            fPInputs = keras_model.input_names
        else:
            fPInputs = [x.name for x in keras_model.inputs]

        fPInputShape = (
            keras_model.input_shape if isinstance(keras_model.input_shape, list) else [keras_model.input_shape]
        )
        fPInputDType = []
        for idx in range(len(keras_model.inputs)):
            dtype = keras_model.inputs[idx].dtype.__str__()
            if dtype == "float32":
                fPInputDType.append(dtype)
            else:
                fPInputDType.append(dtype[9:-2])

        if len(fPInputShape) == 1:
            inputName = fPInputs[0]
            inputDType = SOFIE.ConvertStringToType(fPInputDType[0])
            # convert ot a list to modify batch size
            inputShape = list(fPInputShape[0])
            # set the batch size in case of -1 or None as first value
            if inputShape[0] is None or inputShape[0] <= 0:
                inputShape[0] = batch_size
            rmodel.AddInputTensorInfo(inputName, inputDType, inputShape)
            rmodel.AddInputTensorName(inputName)
        else:
            # Iterating through multiple input tensors
            for inputName, inputDType, inputShapeTuple in zip(fPInputs, fPInputDType, fPInputShape):
                inputDType = SOFIE.ConvertStringToType(inputDType)
                inputShape = list(inputShapeTuple)
                if inputShape[0] is None or inputShape[0] <= 0:
                    inputShape[0] = batch_size
                rmodel.AddInputTensorInfo(inputName, inputDType, inputShape)
                rmodel.AddInputTensorName(inputName)

        # Adding OutputTensorInfos
        outputNames = []
        if keras_version < "2.16" or is_functional_model:
            for layerName in keras_model.output_names:
                final_layer = keras_model.get_layer(layerName)
                output_layer_name = final_layer.output.name
                outputNames.append(output_layer_name)
        else:
            output_layer_name = "tensor_output_" + keras_model.layers[-1].name
            outputNames.append(output_layer_name)

        rmodel.AddOutputTensorNameList(outputNames)
        return rmodel



## \file
## \ingroup tutorial_vecops
## \notebook -nodraw
## In this tutorial we learn how combinations of RVecs can be built.
##
## \macro_code
## \macro_output
##
## \date August 2018
## \author Stefan Wunsch

import ROOT
from ROOT.VecOps import Take, Combinations

# RVec can be sorted in Python with the inbuilt sorting function because
# PyROOT implements a Python iterator
v1 = ROOT.RVecD(3)
v1[0], v1[1], v1[2] = 1, 2, 3
v2 = ROOT.RVecD(2)
v2[0], v2[1] = -4, -5

# To get the indices, which result in all combinations, you can call the
# following helper.
# Note that you can also pass the size of the vectors directly.
idx = Combinations(v1, v2)

# Next, the respective elements can be taken via the computed indices.
c1 = Take(v1, idx[0])
c2 = Take(v2, idx[1])

# Finally, you can perform any set of operations conveniently.
v3 = c1 * c2

print("Combinations of {} and {}:".format(v1, v2))
for i in range(len(v3)):
    print("{} * {} = {}".format(c1[i], c2[i], v3[i]))
print

# However, if you want to compute operations on unique combinations of a
# single RVec, you can perform this as follows.

# Get the indices of unique triples for the given vector.
v4 = ROOT.RVecD(4)
v4[0], v4[1], v4[2], v4[3] = 1, 2, 3, 4
idx2 = Combinations(v4, 3)

# Take the elements and compute any operation on the returned collections.
c3 = Take(v4, idx2[0])
c4 = Take(v4, idx2[1])
c5 = Take(v4, idx2[2])

v5 = c3 * c4 * c5

print("Unique triples of {}:".format(v4))
for i in range(len(v5)):
    print("{} * {} * {} = {}".format(c3[i], c4[i], c5[i], v5[i]))



## \file
## \ingroup tutorial_g3d_legacy
## Before executing this macro, the file geometry.C must have been executed
##
## \macro_code
##
## \author Wim Lavrijsen

import ROOT

ROOT.gBenchmark.Start( 'geometry' )
na = ROOT.TFile( 'na49.root', 'RECREATE' )
n49 = ROOT.gROOT.FindObject( 'na49' )
n49.Write()
na.Write()
na.Close()
ROOT.gBenchmark.Show( 'geometry' )




# Author: Silia Taider CERN  10/2025

################################################################################
# Copyright (C) 1995-2025, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################
from __future__ import annotations

from typing import Any


def _get_axis(self, axis):
    return getattr(self, f"Get{['X', 'Y', 'Z'][axis]}axis")()


def _get_axis_len(self, axis, flow=False):
    return _get_axis(self, axis).GetNbins() + (2 if flow else 0)


def _underflow(hist: Any, axis: int) -> int:
    return 0


def _overflow(hist: Any, axis: int) -> int:
    return _get_axis(hist, axis).GetNbins() + 1


class _loc:
    """
    Represents a location-based index for histograms, returning the bin corresponding
    to a specified value on a given axis. Supports addition and subtraction to shift
    the computed bin by an integer offset.

    Example:
        v = h[loc(b) + 1]  # Returns the bin above the one containing the value `b`
    """

    def __init__(self, value: float) -> None:
        self.value = value
        self.offset = 0

    def __add__(self, other: int) -> _loc:
        if isinstance(other, int):
            self.offset += other
            return self
        raise TypeError(f"Unsupported type for addition: {type(other).__name__}. Expected an integer.")

    def __sub__(self, other: int) -> _loc:
        if isinstance(other, int):
            self.offset -= other
            return self
        raise TypeError(f"Unsupported type for substraction: {type(other).__name__}. Expected an integer.")

    def __call__(self, hist: Any, axis: int) -> int:
        return _get_axis(hist, axis).FindBin(self.value) + self.offset


class _rebin:
    """
    Represents a rebinning operation for histograms, where bins are grouped together
    by the factor ngroup.

    Example:
        h_rebinned = h[::ROOT.uhi.rebin(2)]  # Rebin the histogram with a grouping factor of 2
    """

    def __init__(self, ngroup):
        self.ngroup = ngroup

    def __call__(self, hist):
        rebin_methods = {1: "Rebin", 2: "Rebin2D", 3: "Rebin3D"}
        rebin_method = rebin_methods.get(hist.GetDimension())
        rebin_method = getattr(hist, rebin_method)
        return rebin_method(*self.ngroup, newname=hist.GetName())


def _sum(hist, axis, args=None):
    """
    Represents a summation operation for histograms, which either computes the integral (1D histograms)
    or projects the histogram along specified axes (projection is only for 2D and 3D histograms).

    Example:
        ans = h[0:len:ROOT.uhi.sum]  # Compute the integral for a 1D histogram excluding flow bins
        ans_2 = h[::ROOT.uhi.sum, ::ROOT.uhi.sum]  # Compute the integral for a 2D histogram including flow bins
        h_projected = h[:, ::ROOT.uhi.sum]  # Project the Y axis for a 2D histogram
        h_projected = h[:, :, ::ROOT.uhi.sum]  # Project the Z axis for a 3D histogram
    """
    dim = hist.GetDimension()

    def _invalid_axis(axis, dim):
        raise ValueError(f"Invalid axis {axis} for {dim}D histogram")

    if isinstance(axis, int):
        axis = (axis,)
    if dim == 1:
        return hist.Integral(*args) if axis == (0,) else _invalid_axis(axis, dim)
    if dim == 2:
        if axis == (0,):
            return hist.ProjectionY()
        elif axis == (1,):
            return hist.ProjectionX()
        elif axis == (0, 1):
            return hist.Integral()
        else:
            return _invalid_axis(axis, dim)
    if dim == 3:
        # It is not possible from the interface to specify the options "xy", "yz", "xz"
        project_map = {
            (0,): "zy",
            (1,): "zx",
            (2,): "yx",
            (0, 1): "z",
            (0, 2): "y",
            (1, 2): "x",
        }
        if axis == (0, 1, 2):
            return hist.Integral()
        return hist.Project3D(project_map[axis]) if axis in project_map else _invalid_axis(axis, dim)
    raise NotImplementedError(f"Summing not implemented for {dim}D histograms")



### \file
### \ingroup tutorial_ml
### \notebook -nodraw
###
### \macro_code
### \macro_output
### \author Dante Niewenhuis

##################################################
# This tutorial shows the usage of filters and vectors
# when using the ROOT ML dataloader
##################################################

import ROOT

tree_name = "test_tree"
file_name = ROOT.gROOT.GetTutorialDir().Data() + "/machine_learning/ml_dataloader_filters_vectors_hvector.root"

batch_size = 5  # Defines the size of the returned batches

rdataframe = ROOT.RDataFrame(tree_name, file_name)

# Define filters, filters must be named
filteredrdf = (
    rdataframe.Filter("f1 > 30", "first_filter").Filter("f2 < 70", "second_filter").Filter("f3==true", "third_filter")
)

max_vec_sizes = {"f4": 3, "f5": 2, "f6": 1}

dl = ROOT.Experimental.ML.RDataLoader(
    filteredrdf,
    batch_size,
    max_vec_sizes=max_vec_sizes,
    shuffle=False,
)

ds_train, ds_validation = dl.train_test_split(test_size=0.3)

print(f"Columns: {ds_train.columns}")

for i, b in enumerate(ds_train.as_numpy()):
    print(f"Training batch {i} => {b.shape}")

for i, b in enumerate(ds_validation.as_numpy()):
    print(f"Validation batch {i} => {b.shape}")



# Authors:
# * Lorenzo Moneta 04/2022
# * Harshal Shende 04/2022

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


def _kwargs_to_tmva_cmdargs(*args, **kwargs):
    """Helper function to check kwargs with keys that correspond to a function that creates TmvaCmdArg."""

    def getter(k, v):
        # helper function to get CmdArg attribute from `TMVA`
        # Parameters:
        # k: key of the kwarg
        # v: value of the kwarg

        try:
            if isinstance(v, bool):
                return str(k) if v else "!{}".format(k)
            else:
                return "{k}={v}".format(k=k,v=v)
        except:
            raise AttributeError("Unsupported Type passed")

    if kwargs:
        cmdOpt = ""
        first = True
        for k, v in kwargs.items():
            if not first:
                cmdOpt += ":"
            cmdOpt += getter(k, v)
            first = False

        args = args + (cmdOpt,)

    return args, {}


def cpp_signature(sig):
    """Decorator to set the `_cpp_signature` attribute of a function.
    This information can be used to generate the documentation.
    """

    def decorator(func):
        func._cpp_signature = sig
        return func

    return decorator



## \file
## \ingroup tutorial_gui
## A Simple GUI Example
##
## \macro_code
##
## \author Wim Lavrijsen
from __future__ import print_function

import os, sys, ROOT

def pygaus( x, par ):
   import math
   if (par[2] != 0.0):
      arg1 = (x[0]-par[1])/par[2]
      arg2 = (0.01*0.39894228)/par[2]
      arg3 = par[0]/(1+par[3])

      gauss = arg3*arg2*math.exp(-0.5*arg1*arg1)
   else:
      print('returning 0')
      gauss = 0.
   return gauss

tpygaus = ROOT.TF1( 'pygaus', pygaus, -4, 4, 4 )
tpygaus.SetParameters( 1., 0., 1. )

def MyDraw():
   btn = ROOT.BindObject( ROOT.gTQSender, ROOT.TGTextButton )
   if btn.WidgetId() == 10:
      global tpygaus, window
      tpygaus.Draw()
      ROOT.gPad.Update()

m = ROOT.TPyDispatcher( MyDraw )


class pMainFrame( ROOT.TGMainFrame ):
   def __init__( self, parent, width, height ):
       ROOT.TGMainFrame.__init__( self, parent, width, height )

       self.Canvas    = ROOT.TRootEmbeddedCanvas( 'Canvas', self, 200, 200 )
       self.AddFrame( self.Canvas, ROOT.TGLayoutHints(ROOT.kLHintsExpandX | ROOT.kLHintsExpandY) )
       self.ButtonsFrame = ROOT.TGHorizontalFrame( self, 200, 40 )

       self.DrawButton   = ROOT.TGTextButton( self.ButtonsFrame, '&Draw', 10 )
       self.DrawButton.Connect( 'Clicked()', "TPyDispatcher", m, 'Dispatch()' )
       self.ButtonsFrame.AddFrame( self.DrawButton, ROOT.TGLayoutHints() )

       self.ExitButton   = ROOT.TGTextButton( self.ButtonsFrame, '&Exit', 20 )
       self.ExitButton.SetCommand( 'TPython::Exec( "raise SystemExit" )' )
       self.ButtonsFrame.AddFrame( self.ExitButton, ROOT.TGLayoutHints() )

       self.AddFrame( self.ButtonsFrame, ROOT.TGLayoutHints() )

       self.SetWindowName( 'My first GUI' )
       self.MapSubwindows()
       self.Resize( self.GetDefaultSize() )
       self.MapWindow()

   def __del__(self):
       self.Cleanup()


if __name__ == '__main__':
   window = pMainFrame( ROOT.gClient.GetRoot(), 200, 200 )



## \ingroup tutorial_roofit_main
## \notebook
##
## 'VALIDATION AND MC STUDIES' RooFit tutorial macro #803
##
## RooMCStudy: Using the randomizer and profile likelihood add-on models
##
## \macro_image
## \macro_code
## \macro_output
##
## \date February 2018
## \author Clemens Lange


import ROOT


# Create model
# -----------------------

# Simulation of signal and background of top quark decaying into
# 3 jets with background

# Observable
mjjj = ROOT.RooRealVar("mjjj", "m(3jet) (GeV)", 100, 85.0, 350.0)

# Signal component (Gaussian)
mtop = ROOT.RooRealVar("mtop", "m(top)", 162)
wtop = ROOT.RooRealVar("wtop", "m(top) resolution", 15.2)
sig = ROOT.RooGaussian("sig", "top signal", mjjj, mtop, wtop)

# Background component (Chebychev)
c0 = ROOT.RooRealVar("c0", "Chebychev coefficient 0", -0.846, -1.0, 1.0)
c1 = ROOT.RooRealVar("c1", "Chebychev coefficient 1", 0.112, -1.0, 1.0)
c2 = ROOT.RooRealVar("c2", "Chebychev coefficient 2", 0.076, -1.0, 1.0)
bkg = ROOT.RooChebychev("bkg", "combinatorial background", mjjj, [c0, c1, c2])

# Composite model
nsig = ROOT.RooRealVar("nsig", "number of signal events", 53, 0, 1e3)
nbkg = ROOT.RooRealVar("nbkg", "number of background events", 103, 0, 5e3)
model = ROOT.RooAddPdf("model", "model", [sig, bkg], [nsig, nbkg])

# Create manager
# ---------------------------

# Configure manager to perform binned extended likelihood fits (Binned=True, Extended=True) on data generated
# with a Poisson fluctuation on Nobs (Extended=True)
mcs = ROOT.RooMCStudy(
    model, {mjjj}, Binned=True, Silence=True, Extended=True, FitOptions={"Extended": True, "PrintEvalErrors": -1}
)

# Customize manager
# ---------------------------------

# Add module that randomizes the summed value of nsig+nbkg
# sampling from a uniform distribution between 0 and 1000
#
# In general one can randomize a single parameter, a
# sum of N parameters, either a uniform or a Gaussian
# distribution. Multiple randomization can be executed
# by a single randomizer module

randModule = ROOT.RooRandomizeParamMCSModule()
randModule.sampleSumUniform({nsig, nbkg}, 50, 500)
mcs.addModule(randModule)

# Add profile likelihood calculation of significance. Redo each
# fit while keeping parameter nsig fixed to zero. For each toy,
# the difference in -log(L) of both fits is stored, well
# a simple significance interpretation of the delta(-logL)
# Dnll = 0.5 sigma^2

sigModule = ROOT.RooDLLSignificanceMCSModule(nsig, 0)
mcs.addModule(sigModule)

# Run manager, make plots
# ---------------------------------------------

# Run 1000 experiments. ROOT.This configuration will generate a fair number
# of (harmless) MINUIT warnings due to the instability of the Chebychev polynomial fit
# at low statistics.
mcs.generateAndFit(500)

# Make some plots
binning = ROOT.RooFit.AutoBinning(40)
dll_vs_ngen = mcs.fitParDataSet().createHistogram("ngen,dll_nullhypo_nsig", binning, binning)
z_vs_ngen = mcs.fitParDataSet().createHistogram("ngen,significance_nullhypo_nsig", binning, binning)
errnsig_vs_ngen = mcs.fitParDataSet().createHistogram("ngen,nsigerr", binning, binning)
errnsig_vs_nsig = mcs.fitParDataSet().createHistogram("nsig,nsigerr", binning, binning)

# Draw plots on canvas
c = ROOT.TCanvas("rf803_mcstudy_addons2", "rf802_mcstudy_addons2", 800, 800)
c.Divide(2, 2)
c.cd(1)
ROOT.gPad.SetLeftMargin(0.15)
dll_vs_ngen.GetYaxis().SetTitleOffset(1.6)
dll_vs_ngen.Draw("box")
c.cd(2)
ROOT.gPad.SetLeftMargin(0.15)
z_vs_ngen.GetYaxis().SetTitleOffset(1.6)
z_vs_ngen.Draw("box")
c.cd(3)
ROOT.gPad.SetLeftMargin(0.15)
errnsig_vs_ngen.GetYaxis().SetTitleOffset(1.6)
errnsig_vs_ngen.Draw("box")
c.cd(4)
ROOT.gPad.SetLeftMargin(0.15)
errnsig_vs_nsig.GetYaxis().SetTitleOffset(1.6)
errnsig_vs_nsig.Draw("box")

c.SaveAs("rf803_mcstudy_addons2.png")

# Make ROOT.RooMCStudy object available on command line after
# macro finishes
ROOT.gDirectory.Add(mcs)



# Author: Stephan Hageboeck, CERN 04/2020

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


from ._utils import _kwargs_to_roocmdargs, cpp_signature


class RooWorkspace(object):
    r"""The RooWorkspace::import function can't be used in PyROOT because `import` is a reserved python keyword.
    For this reason, an alternative with a capitalized name is provided:
    \code{.py}
    workspace.Import(x)
    \endcode
    """

    __cpp_name__ = 'RooWorkspace'

    @cpp_signature(
        "Bool_t RooWorkspace::import(const RooAbsArg& arg,"
        "    const RooCmdArg& arg1={},const RooCmdArg& arg2={},const RooCmdArg& arg3={},"
        "    const RooCmdArg& arg4={},const RooCmdArg& arg5={},const RooCmdArg& arg6={},"
        "    const RooCmdArg& arg7={},const RooCmdArg& arg8={},const RooCmdArg& arg9={}) ;"
    )
    def __init__(self, *args, **kwargs):
        r"""The RooWorkspace constructor is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the constructor.
        """
        # Redefinition of `RooWorkspace` constructor for keyword arguments.
        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        self._init(*args, **kwargs)

    def __getitem__(self, key):
        # To enable accessing objects in the RooWorkspace with dictionary-like syntax.
        # The key is passed to the general `RooWorkspace::obj()` function.
        return self.obj(key)

    def __setitem__(self, key, value):
        # Check if initialization is done with string
        if isinstance(value, str):
            if not self.obj(key):
                parenthesis_index = -1
                for i in range(0, len(value)):
                    if value[i] == "[" or value[i] == "(":
                        parenthesis_index = i
                        break
                # Initializes variables
                if value[parenthesis_index] == "[" and parenthesis_index != -1:
                    expr = key + value
                    self.factory(expr)
                # Initializes functions and p.d.f.s
                elif value[parenthesis_index] == "(" and parenthesis_index != -1:
                    expr = value[0:parenthesis_index] + "::" + key + value[parenthesis_index:]
                    self.factory(expr)
                # Else raises a Syntax error
                else:
                    raise SyntaxError("Invalid syntax")
            else:
                raise RuntimeError(
                    "ERROR importing object named "
                    + key
                    + " another instance with same name already in the workspace and no conflict resolution protocol specified"
                )
        elif isinstance(value, dict):
            import json

            import ROOT

            json_string = json.dumps(value, separators=(",", ":"))
            ROOT.RooJSONFactoryWSTool(self).importJSONElement(key, json_string)
        else:
            raise TypeError("Object of type 'str' or 'dict' expected but " + type(value) + " was given")

    @cpp_signature(
        [
            "Bool_t RooWorkspace::import(const RooAbsArg& arg,"
            "         const RooCmdArg& arg1={},const RooCmdArg& arg2={},const RooCmdArg& arg3={},"
            "         const RooCmdArg& arg4={},const RooCmdArg& arg5={},const RooCmdArg& arg6={},"
            "         const RooCmdArg& arg7={},const RooCmdArg& arg8={},const RooCmdArg& arg9={}) ;",
            "Bool_t RooWorkspace::import(RooAbsData& data,"
            "         const RooCmdArg& arg1={},const RooCmdArg& arg2={},const RooCmdArg& arg3={},"
            "         const RooCmdArg& arg4={},const RooCmdArg& arg5={},const RooCmdArg& arg6={},"
            "         const RooCmdArg& arg7={},const RooCmdArg& arg8={},const RooCmdArg& arg9={}) ;",
            "RooWorkspace::import(const char *fileSpec,"
            "         const RooCmdArg& arg1={},const RooCmdArg& arg2={},const RooCmdArg& arg3={},"
            "         const RooCmdArg& arg4={},const RooCmdArg& arg5={},const RooCmdArg& arg6={},"
            "         const RooCmdArg& arg7={},const RooCmdArg& arg8={},const RooCmdArg& arg9={}) ;",
        ]
    )
    def Import(self, *args, **kwargs):
        r"""
        Support the C++ `import()` as `Import()` in python
        """
        return getattr(self, "import")(*args, **kwargs)

    def __setattr__(self, name, value):
        # Many people pythonized the RooWorkspace themselves, by adding a new
        # attribute `_import` that calls getattr(self, "import") under the
        # hood. However, `_import` is now the reference to the original overload
        # proxy, and resetting it with a wrapper around `import` would cause
        # infinite recursions! We prevent resetting any import-related function
        # here, which results in a clearer error to the user than an infinite
        # call stack involving the internal pythonization code.
        if name in ["_import", "import", "Import"]:
            raise AttributeError('Resetting the "' + name + '" attribute of a RooWorkspace is not allowed!')
        object.__setattr__(self, name, value)

    def _ipython_key_completions_(self):
        r"""
        Support tab completion for `__getitem__`, suggesting all components in
        the workspace.
        """
        return [c.GetName() for c in self.components()]


def RooWorkspace_import(self, *args, **kwargs):
    r"""The RooWorkspace::import function can't be used in PyROOT because `import` is a reserved python keyword.
    So, Import() is used and pythonized with the command argument pythonization.
    The keywords must correspond to the CmdArg of the `import()` function.
    """
    # Redefinition of `RooWorkspace.import()` for keyword arguments.
    args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
    return self._import(*args, **kwargs)


setattr(RooWorkspace, "import", RooWorkspace_import)



## \file
## \ingroup tutorial_hist
## \notebook
## Example of macro illustrating how to superimpose two histograms
## with different scales in the "same" pad.
## Inspired by work of Rene Brun.
##
## \macro_image
## \macro_code
##
## \author Alberto Ferro

import ROOT
import numpy as np

c1 = ROOT.TCanvas("c1","hists with different scales",600,400)

ROOT.gStyle.SetOptStat(False)

h1 = ROOT.TH1F("h1","my histogram",100,-3,3)

h1.Fill(np.array([ROOT.gRandom.Gaus(0, 1) for _ in range(10000)]))

h1.Draw()
c1.Update()

hint1 = ROOT.TH1F("hint1","h1 bins integral",100,-3,3)

sum = 0
for i in range(1,101) :
   sum += h1.GetBinContent(i)
   hint1.SetBinContent(i,sum)

rightmax = 1.1*hint1.GetMaximum()
scale = ROOT.gPad.GetUymax()/rightmax
hint1.SetLineColor("kRed")
hint1.Scale(scale)
hint1.Draw("same")

axis = ROOT.TGaxis(ROOT.gPad.GetUxmax(),ROOT.gPad.GetUymin(),
      ROOT.gPad.GetUxmax(), ROOT.gPad.GetUymax(),0,rightmax,510,"+L")
axis.SetLineColor("kRed")
axis.SetLabelColor("kRed")
axis.Draw()



## \file
## \ingroup tutorial_roofit_main
## \notebook
## 'ORGANIZATION AND SIMULTANEOUS FITS' RooFit tutorial macro #512
##
## Illustration of operator expressions and expression-based
## basic p.d.f.s in the workspace factory syntax
##
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C version)


import ROOT


w = ROOT.RooWorkspace("w")

# You can define typedefs for even shorter construction semantics
w.factory("$Typedef(Gaussian,Gaus)")
w.factory("$Typedef(Chebychev,Cheby)")

# Operator pdf examples
# ------------------------------------------------

# PDF addition is done with SUM (coef1*pdf1,pdf2)
w.factory("SUM::summodel( f[0,1]*Gaussian::gx(x[-10,10],m[0],1.0), Chebychev::ch(x,{0.1,0.2,-0.3}) )")

# Extended PDF addition is done with SUM (yield1*pdf1,yield2*pdf2)
w.factory("SUM::extsummodel( Nsig[0,1000]*gx, Nbkg[0,1000]*ch )")

# PDF multiplication is done with PROD ( pdf1, pdf2 )
w.factory("PROD::gxz( gx, Gaussian::gz(z[-10,10],0,1) )")

# Conditional p.d.f multiplication is done with PROD ( pdf1|obs, pdf2 )
w.factory("Gaussian::gy( y[-10,10], x, 1.0 )")
w.factory("PROD::gxycond( gy|x, gx )")

# Convolution (numeric/ fft) is done with NCONV/FCONV (obs,pdf1,pdf2)
w.factory("FCONV::lxg( x, Gaussian::g(x,mg[0],1), Landau::lc(x,0,1) )")

# Simultaneous p.d.f.s are constructed with SIMUL( index, state1=pdf1,
# state2=pdf2,...)
w.factory("SIMUL::smodel( c[A=0,B=1], A=Gaussian::gs(x,m,s[1.0, 0.01, 10.0]), B=Landau::ls(x,0,1) )")

# Operator function examples
# ---------------------------------------------------

# Function multiplication is done with prod (func1, func2,...)
w.factory("prod::uv(u[10],v[10])")

# Function addition is done with sum(func1,func2)
w.factory("sum::uv2(u,v)")

# Lagrangian morphing function for the example shown in rf711_lagrangianmorph
infilename = ROOT.gROOT.GetTutorialDir().Data() + "/roofit/roofit/input_histos_rf_lagrangianmorph.root"
w.factory(
    "lagrangianmorph::morph($observableName('pTV'),$fileName('"
    + infilename
    + "'),$couplings({cHq3[0,1],SM[1]}),$NewPhysics(cHq3=1,SM=0),$folders({'SM_NPsq0','cHq3_NPsq1','cHq3_NPsq2'}))"
)

# Taylor expansion is done with taylorexpand(func,{var1,var2,...},val,order)
w.factory("taylorexpand::te(expr::poly('x^4+5*x^3+2*x^2+x+1',x),{x},0.0,2)")


# Interpreted and compiled expression based pdfs
# ---------------------------------------------------------------------------------------------------

# Create a ROOT.RooGenericPdf interpreted p.d.f. You can use single quotes
# to pass the expression string argument
w.factory("EXPR::G('x*x+1',x)")

# Create a custom compiled p.d.f similar to the above interpreted p.d.f.
# The code required to make self p.d.f. is automatically embedded in
# the workspace
w.factory("CEXPR::GC('x*x+a',{x,a[1]})")

# Compiled and interpreted functions (rather than p.d.f.s) can be made with the lower case
# 'expr' and 'cexpr' types

# Print workspace contents
w.Print()

# Make workspace visible on command line
ROOT.gDirectory.Add(w)



## \file
## \ingroup tutorial_graphs
## \notebook
## \preview Create, draw and fit a TGraph2DErrors. See the [TGraph2DErrors documentation](https://root.cern/doc/master/classTGraph2DErrors.html)
##
## \macro_image
## \macro_code
## \author Olivier Couet, Jamie Gooding

from ctypes import c_double

import ROOT

c1 = ROOT.TCanvas("c1")

e = 0.3
nd = 500

# To generate some random data to put into the graph
r = ROOT.TRandom()
f2 = ROOT.TF2("f2", "1000*(([0]*sin(x)/x)*([1]*sin(y)/y))+200", -6, 6, -6, 6)
f2.SetParameters(1, 1)

dte = ROOT.TGraph2DErrors(nd)

# Fill the 2D graph. It was created only specifying the number of points, so all
# elements are empty. We now "fill" the values and errors with SetPoint and SetPointError.
# Note that the first point has index zero
x = c_double()
y = c_double()
zmax = 0
for i in range(nd):
    f2.GetRandom2(x, y)
    rnd = r.Uniform(-e, e)  # Generate a random number in [-e,e]
    z = f2.Eval(x, y) * (1 + rnd)
    if z > zmax:
        zmax = z
    dte.SetPoint(i, x, y, z)
    ex = 0.05 * r.Rndm()
    ey = 0.05 * r.Rndm()
    ez = abs(z * rnd)
    dte.SetPointError(i, ex, ey, ez)

# If the fit is not needed, just draw dte here and skip the lines below
# dte.Draw("A p0")

# To do the fit we use a function, in this example the same f2 from above
f2.SetParameters(0.5, 1.5)
dte.Fit(f2)
fit2 = dte.FindObject("f2")
fit2.SetTitle("Minuit fit result on the Graph2DErrors points")
fit2.SetMaximum(zmax)
ROOT.gStyle.SetHistTopMargin(0)
fit2.SetLineColor(1)
fit2.SetLineWidth(1)
fit2.Draw("surf1")
dte.Draw("same p0")



## \file
## \ingroup tutorial_dataframe
## \notebook -nodraw
## Basic usage of RDataFrame from python.
##
## This tutorial illustrates the basic features of the RDataFrame class,
## a utility which allows to interact with data stored in TTrees following
## a functional-chain like approach.
##
## \macro_code
## \macro_output
##
## \date May 2017
## \author Danilo Piparo (CERN)

import ROOT

def fill_tree(treeName, fileName):
    """A simple helper function to fill a test tree: this makes the example stand-alone."""
    df = ROOT.RDataFrame(10)
    df.Define("b1", "static_cast<double>(rdfentry_)")\
      .Define("b2", "static_cast<int>(rdfentry_ * rdfentry_)").Snapshot(treeName, fileName)

# We prepare an input tree to run on
fileName = "df001_introduction_py.root"
treeName = "myTree"
fill_tree(treeName, fileName)


# We read the tree from the file and create a RDataFrame, a class that
# allows us to interact with the data contained in the tree.
d = ROOT.RDataFrame(treeName, fileName)

# Operations on the dataframe
# We now review some *actions* which can be performed on the data frame.
# Actions can be divided into instant actions (e. g. Foreach()) and lazy
# actions (e. g. Count()), depending on whether they trigger the event 
# loop immediately or only when one of the results is accessed for the 
# first time. Actions that return "something" either return their result 
# wrapped in a RResultPtr or in a RDataFrame.
# But first of all, let us we define now our cut-flow with two strings.
# Filters can be expressed as strings. The content must be C++ code. The
# name of the variables must be the name of the branches. The code is
# just-in-time compiled.
cutb1 = 'b1 < 5.'
cutb1b2 = 'b2 % 2 && b1 < 4.'

# `Count` action
# The `Count` allows to retrieve the number of the entries that passed the
# filters. Here we show how the automatic selection of the column kicks
# in in case the user specifies none.
entries1 = d.Filter(cutb1) \
            .Filter(cutb1b2) \
            .Count();

print('{} entries passed all filters'.format(entries1.GetValue()))

entries2 = d.Filter("b1 < 5.").Count();
print('{} entries passed all filters'.format(entries2.GetValue()))

# `Min`, `Max` and `Mean` actions
# These actions allow to retrieve statistical information about the entries
# passing the cuts, if any.
b1b2_cut = d.Filter(cutb1b2)
minVal = b1b2_cut.Min('b1')
maxVal = b1b2_cut.Max('b1')
meanVal = b1b2_cut.Mean('b1')
nonDefmeanVal = b1b2_cut.Mean("b2")
print('The mean is always included between the min and the max: {0} <= {1} <= {2}'.format(minVal.GetValue(), meanVal.GetValue(), maxVal.GetValue()))

# `Histo1D` action
# The `Histo1D` action allows to fill an histogram. It returns a TH1F filled
# with values of the column that passed the filters. For the most common
# types, the type of the values stored in the column is automatically
# guessed.
hist = d.Filter(cutb1).Histo1D('b1')
print('Filled h {0} times, mean: {1}'.format(hist.GetEntries(), hist.GetMean()))

# Express your chain of operations with clarity!
# We are discussing an example here but it is not hard to imagine much more
# complex pipelines of actions acting on data. Those might require code
# which is well organised, for example allowing to conditionally add filters
# or again to clearly separate filters and actions without the need of
# writing the entire pipeline on one line. This can be easily achieved.
# We'll show this re-working the `Count` example:
cutb1_result = d.Filter(cutb1);
cutb1b2_result = d.Filter(cutb1b2);
cutb1_cutb1b2_result = cutb1_result.Filter(cutb1b2)

# Now we want to count:
evts_cutb1_result = cutb1_result.Count()
evts_cutb1b2_result = cutb1b2_result.Count()
evts_cutb1_cutb1b2_result = cutb1_cutb1b2_result.Count()

print('Events passing cutb1: {}'.format(evts_cutb1_result.GetValue()))
print('Events passing cutb1b2: {}'.format(evts_cutb1b2_result.GetValue()))
print('Events passing both: {}'.format(evts_cutb1_cutb1b2_result.GetValue()))

# Calculating quantities starting from existing columns
# Often, operations need to be carried out on quantities calculated starting
# from the ones present in the columns. We'll create in this example a third
# column, the values of which are the sum of the *b1* and *b2* ones, entry by
# entry. The way in which the new quantity is defined is via a callable.
# It is important to note two aspects at this point:
# - The value is created on the fly only if the entry passed the existing
# filters.
# - The newly created column behaves as the one present on the file on disk.
# - The operation creates a new value, without modifying anything. De facto,
# this is like having a general container at disposal able to accommodate
# any value of any type.
# Let's dive in an example:
entries_sum = d.Define('sum', 'b2 + b1') \
               .Filter('sum > 4.2') \
               .Count()
print(entries_sum.GetValue())




# Author: Dante Niewenhuis, VU Amsterdam 07/2023
# Author: Kristupas Pranckietis, Vilnius University 05/2024
# Author: Nopphakorn Subsa-Ard, King Mongkut's University of Technology Thonburi (KMUTT) (TH) 08/2024
# Author: Vincenzo Eduardo Padulano, CERN 10/2024
# Author: Martin Føll, University of Oslo (UiO) & CERN 01/2026
# Author: Silia Taider, CERN 02/2026

################################################################################
# Copyright (C) 1995-2026, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from __future__ import annotations

import atexit
from typing import TYPE_CHECKING, Any, Callable, Tuple

if TYPE_CHECKING:
    import numpy as np
    import tensorflow as tf
    import torch

    import ROOT


class _RDataLoader:
    def get_template(
        self,
        x_rdf: ROOT.RDF.RNode,
        columns: list[str] | None = None,
        max_vec_sizes: dict[str, int] | None = None,
    ) -> Tuple[str, list[int]]:
        """
        Generate a template for the DataLoader based on the given
        RDataFrame and columns.

        Args:
            x_rdf (RNode): RDataFrame or RNode object.
            columns (list[str], optional): Columns that should be loaded.
                                 Defaults to loading all columns
                                 in the given RDataFrame
            max_vec_sizes (dict[str, int], optional):
                                 Mapping from vector column name
                                 to the maximum size of the vector.
                                 Required when using vector based columns.

        Returns:
            Tuple[str, list[int]]: Template string for the DataLoader and list of max vector sizes
        """
        if not columns:
            columns = x_rdf.GetColumnNames()
        if max_vec_sizes is None:
            max_vec_sizes = {}

        template_string = ""

        self.given_columns = []
        self.all_columns = []

        max_vec_sizes_list = []

        for name in columns:
            name_str = str(name)
            self.given_columns.append(name_str)
            column_type = x_rdf.GetColumnType(name_str)
            template_string = f"{template_string}{column_type},"

            if "RVec" in column_type:
                # Add column for each element if column is a vector
                if name_str in max_vec_sizes:
                    max_vec_sizes_list.append(max_vec_sizes[name_str])
                    for i in range(max_vec_sizes[name_str]):
                        self.all_columns.append(f"{name_str}_{i}")

                else:
                    raise ValueError(
                        f"No max size given for feature {name_str}. \
                        Given max sizes: {max_vec_sizes}"
                    )

            else:
                self.all_columns.append(name_str)

        return template_string[:-1], max_vec_sizes_list

    def __init__(
        self,
        rdataframes: ROOT.RDF.RNode | list[ROOT.RDF.RNode] | None = None,
        batch_size: int = 0,
        batches_in_memory: int = 1,
        columns: list[str] | None = None,
        max_vec_sizes: dict[str, int] | None = None,
        vec_padding: int = 0,
        target: str | list[str] | None = None,
        weights: str = "",
        test_size: float = 0.0,
        shuffle: bool = True,
        drop_remainder: bool = True,
        set_seed: int = 0,
        load_eager: bool = False,
        sampling_type: str = "",
        sampling_ratio: float = 1.0,
        replacement: bool = False,
    ) -> None:
        """Wrapper around the C++ DataLoader

        Args:
            rdataframes (ROOT.RDF.RNode | list[ROOT.RDF.RNode] | None):
                RDataFrame or list of RDataFrames to load from.
            batch_size (int):
                Number of entries per batch returned by the generator.
            batches_in_memory (int):
                Approximate number of batches that should be kept in memory at
                the same time. Higher value results in faster loading, but
                also higher memory usage. Defaults to 1.
            columns (list[str] | None):
                Names of columns to load. If not given, all columns are used.
            max_vec_sizes (dict[str, int] | None):
                Mapping from vector column name to the maximum size of the vector.
                Required when using vector based columns.
            vec_padding (int):
                Value used to pad vectors with if the vector is smaller
                than the given max vector length. Defaults to 0.
            target (str | list[str] | None):
                Name or list of names of target column(s).
            weights (str):
                Column used to weight events.
                Can only be used when a target is given.
            test_size (float):
                The ratio of batches being kept for validation.
                Value has to be between 0 and 1. Defaults to 0.0.
            shuffle (bool):
                Batches consist of random events and are shuffled every epoch.
                Defaults to True.
            drop_remainder (bool):
                Drop the remainder of data that is too small to compose full batch.
                Defaults to True.
            set_seed (int):
                For reproducibility: Set the seed for the random number generator used
                to split the dataset into training and validation as well as shuffing of the entries.
                Defaults to 0 which means that the seed is set to the random device.
            load_eager (bool):
                If True, load the full dataset(s) into memory.
                If False, load data lazily in clusters. Defaults to False.
            sampling_type (str):
                Describes the mode of sampling from the minority and majority dataframes.
                Supported values are ``"undersampling"`` and ``"oversampling"``. Requires ``load_eager=True``.
                Defaults to ``""``.
                For 'undersampling' and 'oversampling' it requires a list of exactly two dataframes as input,
                where the dataframe with the most entries is the majority dataframe
                and the dataframe with the fewest entries is the minority dataframe.
            sampling_ratio (float):
                Ratio of minority and majority entries in the resampled dataset.
                Requires ``load_eager=True`` and ``sampling_type="undersampling"`` or ``"oversampling"``. Defaults to 1.0.
            replacement (bool):
                Whether the sampling is with (True) or without (False) replacement.
                Requires ``load_eager=True`` and ``sampling_type="undersampling"``. Defaults to False.
        """

        from ROOT import RDF

        if rdataframes is None:
            rdataframes = []
        if columns is None:
            columns = []
        if max_vec_sizes is None:
            max_vec_sizes = {}
        if target is None or target == "":
            target = []

        if not hasattr(rdataframes, "__iter__"):
            rdataframes = [rdataframes]
        self.noded_rdfs = [RDF.AsRNode(rdf) for rdf in rdataframes]

        if isinstance(target, str):
            target = [target]

        self.target_columns = target
        self.weights_column = weights

        template, max_vec_sizes_list = self.get_template(self.noded_rdfs[0], columns, max_vec_sizes)

        self.num_columns = len(self.all_columns)
        self.batch_size = batch_size

        # Handle target
        self.target_given = len(self.target_columns) > 0
        self.weights_given = len(self.weights_column) > 0
        if self.target_given:
            for target in self.target_columns:
                if target not in self.all_columns:
                    raise ValueError(
                        f"Provided target not in given columns: \ntarget => \
                            {target}\ncolumns => {self.all_columns}"
                    )

            self.target_indices = [self.all_columns.index(target) for target in self.target_columns]

            # Handle weights
            if self.weights_given:
                if weights in self.all_columns:
                    self.weights_index = self.all_columns.index(self.weights_column)
                    self.train_indices = [
                        c for c in range(len(self.all_columns)) if c not in self.target_indices + [self.weights_index]
                    ]
                else:
                    raise ValueError(
                        f"Provided weights not in given columns: \nweights => \
                            {weights}\ncolumns => {self.all_columns}"
                    )
            else:
                self.train_indices = [c for c in range(len(self.all_columns)) if c not in self.target_indices]

        elif self.weights_given:
            raise ValueError("Weights can only be used when a target is provided")
        else:
            self.train_indices = [c for c in range(len(self.all_columns))]

        self.train_columns = [c for c in self.all_columns if c not in self.target_columns + [self.weights_column]]

        import ROOT

        # The DataLoader will create a separate C++ thread for I/O.
        # Enable thread safety in ROOT from here, to make sure there is no
        # interference between the main Python thread (which might call into
        # cling via cppyy) and the I/O thread.
        ROOT.EnableThreadSafety()

        self.engine = ROOT.Experimental.Internal.ML.RDataLoaderEngine(template)(
            self.noded_rdfs,
            batch_size,
            batches_in_memory,
            self.given_columns,
            max_vec_sizes_list,
            vec_padding,
            test_size,
            shuffle,
            drop_remainder,
            set_seed,
            load_eager,
            sampling_type,
            sampling_ratio,
            replacement,
        )

        atexit.register(self.DeActivate)

    @property
    def isActive(self):
        return self.engine.IsActive()

    def isTrainingActive(self):
        return self.engine.IsTrainingActive()

    def isValidationActive(self):
        return self.engine.IsValidationActive()

    def Activate(self):
        """Initialize the generator to be used for a loop, this spawns the loading thread"""
        self.engine.Activate()

    def DeActivate(self):
        """Deactivate the generator"""
        self.engine.DeActivate()

    def ActivateTrainingEpoch(self):
        """Activate the training epoch of the generator"""
        self.engine.ActivateTrainingEpoch()

    def ActivateValidationEpoch(self):
        """Activate the validation epoch of the generator"""
        self.engine.ActivateValidationEpoch()

    def DeActivateTrainingEpoch(self):
        """Deactivate the training epoch of the generator"""
        self.engine.DeActivateTrainingEpoch()

    def DeActivateValidationEpoch(self):
        """Deactivate the validation epoch of the generator"""
        self.engine.DeActivateValidationEpoch()

    def CreateTrainBatches(self):
        """Create the first training batches from the first cluster"""
        self.engine.CreateTrainBatches()

    def CreateValidationBatches(self):
        """Create the first validation batches from the first cluster"""
        self.engine.CreateValidationBatches()

    @property
    def num_training_batches(self) -> int:
        return self.engine.NumberOfTrainingBatches()

    @property
    def num_validation_batches(self) -> int:
        return self.engine.NumberOfValidationBatches()

    @property
    def train_remainder_rows(self) -> int:
        return self.engine.TrainRemainderRows()

    @property
    def val_remainder_rows(self) -> int:
        return self.engine.ValidationRemainderRows()

    def GetSample(self):
        """
        Return a sample of data that has the same size and types as the actual
        result. This sample can be used to define the shape and size of the
        output

        Returns:
            np.ndarray: data sample
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError("Failed to import numpy needed for the ML dataloader")

        # Split the target and weight
        if not self.target_given:
            return np.zeros((self.batch_size, self.num_columns))

        if not self.weights_given:
            if len(self.target_indices) == 1:
                return np.zeros((self.batch_size, self.num_columns - 1)), np.zeros((self.batch_size)).reshape(-1, 1)

            return np.zeros((self.batch_size, self.num_columns - 1)), np.zeros(
                (self.batch_size, len(self.target_indices))
            )

        if len(self.target_indices) == 1:
            return (
                np.zeros((self.batch_size, self.num_columns - 2)),
                np.zeros((self.batch_size)).reshape(-1, 1),
                np.zeros((self.batch_size)).reshape(-1, 1),
            )

        return (
            np.zeros((self.batch_size, self.num_columns - 2)),
            np.zeros((self.batch_size, len(self.target_indices))),
            np.zeros((self.batch_size)).reshape(-1, 1),
        )

    def ConvertBatchToNumpy(self, batch) -> np.ndarray:
        """Convert a RTensor into a NumPy array

        Args:
            batch (RTensor): Batch returned from the DataLoader

        Returns:
            np.ndarray: converted batch
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError("Failed to import numpy needed for the ML dataloader")

        data = batch.GetData()
        batch_size, num_columns = tuple(batch.GetShape())

        data.reshape((batch_size * num_columns,))

        return_data = np.asarray(data).reshape(batch_size, num_columns)

        # Splice target column from the data if target is given
        if self.target_given:
            train_data = return_data[:, self.train_indices]
            target_data = return_data[:, self.target_indices]

            # Splice weight column from the data if weight is given
            if self.weights_given:
                weights_data = return_data[:, self.weights_index]

                if len(self.target_indices) == 1:
                    return train_data, target_data.reshape(-1, 1), weights_data.reshape(-1, 1)

                return train_data, target_data, weights_data.reshape(-1, 1)

            if len(self.target_indices) == 1:
                return train_data, target_data.reshape(-1, 1)

            return train_data, target_data

        return return_data

    def ConvertBatchToPyTorch(self, batch: Any, device=None) -> torch.Tensor:
        """Convert a RTensor into a PyTorch tensor

        Args:
            batch (RTensor): Batch returned from the DataLoader

        Returns:
            torch.Tensor: converted batch
        """
        import numpy as np
        import torch

        data = batch.GetData()
        batch_size, num_columns = tuple(batch.GetShape())

        data.reshape((batch_size * num_columns,))

        return_data = torch.as_tensor(np.asarray(data), device=device).reshape(batch_size, num_columns)

        # Splice target column from the data if target is given
        if self.target_given:
            train_data = return_data[:, self.train_indices]
            target_data = return_data[:, self.target_indices]

            # Splice weight column from the data if weight is given
            if self.weights_given:
                weights_data = return_data[:, self.weights_index]

                if len(self.target_indices) == 1:
                    return train_data, target_data.reshape(-1, 1), weights_data.reshape(-1, 1)

                return train_data, target_data, weights_data.reshape(-1, 1)

            if len(self.target_indices) == 1:
                return train_data, target_data.reshape(-1, 1)

            return train_data, target_data

        return return_data

    def ConvertBatchToTF(self, batch: Any) -> Any:
        """
        Convert a RTensor into a TensorFlow tensor

        Args:
            batch (RTensor): Batch returned from the DataLoader

        Returns:
            tensorflow.Tensor: converted batch
        """
        import tensorflow as tf

        data = batch.GetData()
        batch_size, num_columns = tuple(batch.GetShape())

        data.reshape((batch_size * num_columns,))

        return_data = tf.constant(data, shape=(batch_size, num_columns))

        if batch_size != self.batch_size:
            return_data = tf.pad(return_data, tf.constant([[0, self.batch_size - batch_size], [0, 0]]))

        # Splice target column from the data if weight is given
        if self.target_given:
            train_data = tf.gather(return_data, indices=self.train_indices, axis=1)
            target_data = tf.gather(return_data, indices=self.target_indices, axis=1)

            # Splice weight column from the data if weight is given
            if self.weights_given:
                weights_data = tf.gather(return_data, indices=[self.weights_index], axis=1)

                return train_data, target_data, weights_data

            return train_data, target_data

        return return_data

    # Return a batch when available
    def GetTrainBatch(self) -> Any:
        """Return the next training batch of data from the given RDataFrame

        Returns:
            (np.ndarray): Batch of data of size.
        """

        batch = self.engine.GetTrainBatch()
        return batch if (batch and batch.GetSize() > 0) else None

    def GetValidationBatch(self) -> Any:
        """Return the next training batch of data from the given RDataFrame

        Returns:
            (np.ndarray): Batch of data of size.
        """

        batch = self.engine.GetValidationBatch()
        return batch if (batch and batch.GetSize() > 0) else None


# context managers for the loading thread
class _TrainingEpochContext:
    def __init__(self, internal: _RDataLoader):
        self._internal = internal
        # init loading thread
        internal.Activate()
        internal.CreateTrainBatches()

    def __enter__(self):
        self._internal.ActivateTrainingEpoch()
        return self

    def __exit__(self, type, value, traceback):
        self._internal.DeActivateTrainingEpoch()


class _ValidationEpochContext:
    def __init__(self, internal: _RDataLoader):
        self._internal = internal
        internal.Activate()
        internal.CreateValidationBatches()

    def __enter__(self):
        self._internal.ActivateValidationEpoch()
        return self

    def __exit__(self, type, value, traceback):
        self._internal.DeActivateValidationEpoch()


# formatted iterator (returned by as_torch / as_numpy / as_tensorflow)
class FormattedLoader:
    """
    Iterable that converts each batch to the requested format.
    Returned by the as_torch / as_numpy / as_tensorflow methods on RDataLoader.
    """

    def __init__(
        self,
        internal: _RDataLoader,
        conversion_fn: Callable,
        is_training: bool,
    ):
        self._internal = internal
        self._conversion_fn = conversion_fn
        self._is_training = is_training

    def _make_gen(self):
        ctx_cls = _TrainingEpochContext if self._is_training else _ValidationEpochContext
        get_batch = self._internal.GetTrainBatch if self._is_training else self._internal.GetValidationBatch

        with ctx_cls(self._internal):
            while True:
                batch = get_batch()
                if batch is None:
                    break
                yield self._conversion_fn(batch)

    def __iter__(self):
        return self._make_gen()


class RDataLoader:
    """
    Entry point for ML batch loading from a ROOT RDataFrame.

    Usage without a validation split::

        train = ROOT.Experimental.ML.RDataLoader(df, batch_size=1000, ...)
        for x, y in train.as_torch():
            ...

    Usage with a validation split::

        dl = ROOT.Experimental.ML.RDataLoader(df, batch_size=1000, ...)
        train, val = dl.train_test_split(test_size=0.2)
        for x, y in train.as_torch():
            ...
        for x, y in val.as_numpy():
            ...
    """

    def __init__(
        self,
        rdataframes: ROOT.RDF.RNode | list[ROOT.RDF.RNode],
        batch_size: int = 64,
        batches_in_memory: int = 10,
        columns: list[str] | None = None,
        max_vec_sizes: dict[str, int] | None = None,
        vec_padding: float = 0.0,
        target: str | list[str] | None = None,
        weights: str = "",
        shuffle: bool = True,
        drop_remainder: bool = True,
        set_seed: int = 0,
        load_eager: bool = False,
        sampling_type: str = "",
        sampling_ratio: float = 1.0,
        replacement: bool = False,
    ) -> None:
        """
        Args:
            rdataframes:
                RDataFrame or list of RDataFrames to load from.
            batch_size:
                Number of entries per batch.
            batches_in_memory:
                Approximate number of batches held in the shuffle buffer at any
                time. Larger values improve shuffle quality across cluster
                boundaries at the cost of higher memory usage. Acts as a soft
                cap: the buffer may temporarily exceed this. Defaults to 10.
            columns:
                Names of columns to load. If not given, all columns are used.
            max_vec_sizes:
                Maximum size per vector column. Required for RVec columns.
            vec_padding:
                Padding value for vectors shorter than their max size. Defaults to 0.
            target:
                Name or list of names of target column(s).
            weights:
                Column to use for event weighting. Requires a target.
            shuffle:
                Whether to shuffle data across cluster boundaries every epoch.
                Defaults to True.
            drop_remainder:
                Drop the last batch if smaller than batch_size. Defaults to True.
            set_seed:
                Seed for the random number generator. 0 means a random seed is
                drawn from the system. Defaults to 0.
            load_eager:
                If True, load the full dataset into memory before training.
                If False (default), load lazily in chunks.
            sampling_type:
                Resampling strategy: "undersampling" or "oversampling".
                Requires load_eager=True and exactly two input dataframes.
            sampling_ratio:
                Ratio of minority to majority entries in the resampled dataset.
                Requires load_eager=True and sampling_type set.
            replacement:
                Whether undersampling is with replacement. Requires load_eager=True
                and sampling_type="undersampling".
        """
        # Store all constructor parameters. The C++ backend (_RDataLoader) is
        # created lazily on the first call to as_torch/as_numpy/as_tensorflow or
        # train_test_split.
        self._params = dict(
            rdataframes=rdataframes,
            batch_size=batch_size,
            batches_in_memory=batches_in_memory,
            columns=columns,
            max_vec_sizes=max_vec_sizes,
            vec_padding=vec_padding,
            target=target,
            weights=weights,
            shuffle=shuffle,
            drop_remainder=drop_remainder,
            set_seed=set_seed,
            load_eager=load_eager,
            sampling_type=sampling_type,  # TODO(staider) consider turning into an enum
            sampling_ratio=sampling_ratio,
            replacement=replacement,
        )
        self._internal: _RDataLoader | None = None
        self._test_size: float | None = None
        self._is_training: bool = True  # default: full dataset treated as training

    @classmethod
    def _from_internal(cls, internal: _RDataLoader, is_training: bool) -> RDataLoader:
        """
        Internal factory that creates a split bound RDataLoader sharing an
        already-constructed C++ backend, train_test_split uses this to return two
        RDataLoader instances that both point at the same _RDataLoader object.
        """
        obj = cls.__new__(cls)
        obj._params = None
        obj._internal = internal
        obj._test_size = None
        obj._is_training = is_training
        return obj

    def _ensure_created(self, test_size: float = 0.0) -> None:
        """
        Construct the C++ backend if not already done.
        """
        if self._internal is not None:
            # Already constructed, guard against accidentally calling with a different split
            if self._params is not None and test_size != self._test_size:
                raise RuntimeError(
                    f"RDataLoader was already initialised with test_size="
                    f"{self._test_size}. Create a new RDataLoader to use a different split."
                )
            return

        self._internal = _RDataLoader(**self._params, test_size=test_size)
        self._test_size = test_size

    def train_test_split(self, test_size: float = 0.2) -> Tuple[RDataLoader, RDataLoader]:
        """
        Partition the dataset into training and validation splits.
        Returns two RDataLoader instances that share the same underlying C++
        backend and can each be iterated independently.
        """
        if not (0.0 < test_size < 1.0):
            raise ValueError(f"test_size must be in (0.0, 1.0), got {test_size}")

        self._ensure_created(test_size)
        return (
            RDataLoader._from_internal(self._internal, is_training=True),
            RDataLoader._from_internal(self._internal, is_training=False),
        )

    def as_numpy(self) -> FormattedLoader:
        """
        Return an iterable that yields batches as NumPy arrays.
        """
        self._ensure_created()
        return FormattedLoader(self._internal, self._internal.ConvertBatchToNumpy, self._is_training)

    def as_torch(self, device: str | torch.device | None = None) -> FormattedLoader:
        """
        Return an iterable that yields batches as PyTorch tensors.

        Args:
            device: If given, the returned tensors are moved to the specified device.
        """
        self._ensure_created()
        conversion_fn = lambda batch: self._internal.ConvertBatchToPyTorch(batch, device)  # noqa: E731
        return FormattedLoader(self._internal, conversion_fn, self._is_training)

    def as_tensorflow(self) -> tf.data.Dataset:
        """
        Return a tf.data.Dataset over batches as TensorFlow tensors.
        """
        import tensorflow as tf

        self._ensure_created()

        batch_size = self._internal.batch_size
        num_train_columns = len(self._internal.train_columns)
        num_target_columns = len(self._internal.target_columns)

        # No target and weights given
        if not self._internal.target_given:
            batch_signature = tf.TensorSpec(shape=(batch_size, num_train_columns), dtype=tf.float32)

        # Target given, no weights given
        elif not self._internal.weights_given:
            batch_signature = (
                tf.TensorSpec(shape=(batch_size, num_train_columns), dtype=tf.float32),
                tf.TensorSpec(shape=(batch_size, num_target_columns), dtype=tf.float32),
            )

        # Target and weights given
        else:
            batch_signature = (
                tf.TensorSpec(shape=(batch_size, num_train_columns), dtype=tf.float32),
                tf.TensorSpec(shape=(batch_size, num_target_columns), dtype=tf.float32),
                tf.TensorSpec(shape=(batch_size, 1), dtype=tf.float32),
            )

        loader = FormattedLoader(self._internal, self._internal.ConvertBatchToTF, self._is_training)
        return tf.data.Dataset.from_generator(lambda: loader, output_signature=batch_signature)

    @property
    def columns(self) -> list[str]:
        """All column names as they appear in each batch tensor."""
        if self._internal is None:
            return self._params["columns"]
        return self._internal.all_columns

    @property
    def train_columns(self) -> list[str]:
        """Feature column names (columns minus target and weights)."""
        if self._internal is None:
            target = self._params["target"] if self._params["target"] is not None else []
            weights = self._params["weights"] if self._params["weights"] is not None else []
            return [col for col in self._params["columns"] if col not in target and col not in weights]
        return self._internal.train_columns

    @property
    def target_columns(self) -> list[str]:
        """Target column names."""
        if self._internal is None:
            return self._params["target"] if self._params["target"] is not None else []
        return self._internal.target_columns

    @property
    def weights_column(self) -> str:
        """Weights column name, or empty string if not set."""
        if self._internal is None:
            return self._params["weights"] if self._params["weights"] is not None else ""
        return self._internal.weights_column

    @property
    def num_batches(self) -> int:
        """Total number of batches in this split for one epoch."""
        if self._internal is None:
            raise RuntimeError(
                "num_batches is available after the first call to "
                "as_torch / as_numpy / as_tensorflow / train_test_split."
            )
        if self._is_training:
            return self._internal.num_training_batches
        return self._internal.num_validation_batches

    @property
    def last_batch_no_of_rows(self) -> int:
        """Number of rows in the last (remainder) batch, 0 if no remainder."""
        if self._internal is None:
            raise RuntimeError(
                "last_batch_no_of_rows is available after the first call to "
                "as_torch / as_numpy / as_tensorflow / train_test_split."
            )
        if self._is_training:
            return self._internal.train_remainder_rows
        return self._internal.val_remainder_rows


def _inject_dataloader_api(parentmodule):
    """
    Inject the public Python API into the ROOT.Experimental.ML namespace.
    Only RDataLoader is part of the public surface.
    """
    for cls in [RDataLoader, FormattedLoader]:
        setattr(parentmodule, cls.__name__, cls)



## \file
## \ingroup tutorial_dataframe
## \notebook -draw
## Process a CSV file with RDataFrame and the CSV data source.
##
## This tutorial illustrates how use the RDataFrame in combination with a
## RDataSource. In this case we use a TCsvDS. This data source allows to read
## a CSV file from a RDataFrame.
## As a result of running this tutorial, we will produce plots of the dimuon
## spectrum starting from a subset of the CMS collision events of Run2010B.
## Dataset Reference:
## McCauley, T. (2014). Dimuon event information derived from the Run2010B
## public Mu dataset. CERN Open Data Portal.
## DOI: [10.7483/OPENDATA.CMS.CB8H.MFFA](http://opendata.cern.ch/record/700).
##
## \macro_code
## \macro_image
##
## \date October 2017
## \author Enric Tejedor (CERN)

import ROOT

# Let's first create a RDF that will read from the remote CSV file.
# The types of the columns will be automatically inferred.
fileUrl = "http://root.cern/files/tutorials/df014_CsvDataSource_MuRun2010B.csv"
df = ROOT.RDF.FromCSV(fileUrl)

# Now we will apply a first filter based on two columns of the CSV,
# and we will define a new column that will contain the invariant mass.
# Note how the new invariant mass column is defined from several other
# columns that already existed in the CSV file.
filteredEvents = df.Filter("Q1 * Q2 == -1") \
                   .Define("m", "sqrt(pow(E1 + E2, 2) - (pow(px1 + px2, 2) + pow(py1 + py2, 2) + pow(pz1 + pz2, 2)))")

# Next we create a histogram to hold the invariant mass values and we draw it.
invMass = filteredEvents.Histo1D(("invMass", "CMS Opendata: #mu#mu mass;#mu#mu mass [GeV];Events", 512, 2, 110), "m")

c = ROOT.TCanvas()
c.SetLogx()
c.SetLogy()
invMass.Draw()
c.SaveAs("df014_invMass.png")

# We will now produce a plot also for the J/Psi particle. We will plot
# on the same canvas the full spectrum and the zoom in on the J/psi particle.
# First we will create the full spectrum histogram from the invariant mass
# column, using a different histogram model than before.
fullSpectrum = filteredEvents.Histo1D(("Spectrum", "Subset of CMS Run 2010B;#mu#mu mass [GeV];Events", 1024, 2, 110), "m")

# Next we will create the histogram for the J/psi particle, applying first
# the corresponding cut.
jpsiLow = 2.95
jpsiHigh = 3.25
jpsiCut = 'm < %s && m > %s' % (jpsiHigh, jpsiLow)
jpsi = filteredEvents.Filter(jpsiCut) \
                     .Histo1D(("jpsi", "Subset of CMS Run 2010B: J/#psi window;#mu#mu mass [GeV];Events", 128, jpsiLow, jpsiHigh), "m")

# Finally we draw the two histograms side by side.
dualCanvas = ROOT.TCanvas("DualCanvas", "DualCanvas", 800, 512)
dualCanvas.Divide(2, 1)
leftPad = dualCanvas.cd(1)
leftPad.SetLogx()
leftPad.SetLogy()
fullSpectrum.Draw("Hist")
dualCanvas.cd(2)
jpsi.SetMarkerStyle(20)
jpsi.Draw("HistP")
dualCanvas.SaveAs("df014_jpsi.png")

print("Saved figures to df014_*.png")



# Authors:
# * Jonas Rembser 06/2021
# * Harshal Shende 06/2021

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


from ._utils import _kwargs_to_roocmdargs, cpp_signature


class RooMsgService(object):
    r"""Some member functions of RooMsgService that take a RooCmdArg as argument also support keyword arguments.
    So far, this applies to RooMsgService::addStream.
    For example, the following code is equivalent in PyROOT:
    \code{.py}
    # Directly passing a RooCmdArg:
    ROOT.RooMsgService.instance().addStream(ROOT.RooFit.DEBUG, ROOT.RooFit.Topic(ROOT.RooFit.Tracing), ROOT.RooFit.ClassName("RooGaussian"))

    # With keyword arguments:
    ROOT.RooMsgService.instance().addStream(ROOT.RooFit.DEBUG, Topic = ROOT.RooFit.Tracing, ClassName = "RooGaussian")
    \endcode"""

    __cpp_name__ = 'RooMsgService'

    @cpp_signature(
        "Int_t RooMsgService::addStream(RooFit::MsgLevel level, const RooCmdArg& arg1={}, const RooCmdArg& arg2={}, const RooCmdArg& arg3={},"
        "    const RooCmdArg& arg4={}, const RooCmdArg& arg5={}, const RooCmdArg& arg6={});"
    )
    def addStream(self, *args, **kwargs):
        r"""The RooMsgService::addStream() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArg of the function.
        """
        # Redefinition of `RooMsgService.addStream` for keyword arguments.
        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        return self._addStream(*args, **kwargs)



# Author: Silia Taider CERN  10/2025

################################################################################
# Copyright (C) 1995-2025, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################
from __future__ import annotations

from contextlib import contextmanager

from .plotting import _values_by_copy
from .tags import _get_axis_len, _overflow, _rebin, _sum, _underflow

"""
Implementation of the indexing component of the UHI
"""


@contextmanager
def _temporarily_disable_add_directory():
    """
    Temporarily disable adding the new created histograms to the list of objects in memory
    """
    import ROOT

    old_status = ROOT.TH1.AddDirectoryStatus()
    ROOT.TH1.AddDirectory(False)
    try:
        yield
    finally:
        ROOT.TH1.AddDirectory(old_status)


def _process_index_for_axis(self, index, axis, is_slice_stop=False):
    """Process an index for a histogram axis handling callables and index shifting."""
    if callable(index):
        # If the index is a `loc`, `underflow`, `overflow`, or `len`
        return _get_axis_len(self, axis) + 1 if index is len else index(self, axis)

    if isinstance(index, int):
        # -1 index returns the last valid bin
        if index == -1:
            return _overflow(self, axis) - 1

        # Shift the indices by 1 to align with the UHI convention,
        # where 0 corresponds to the first bin, unlike ROOT where 0 represents underflow and 1 is the first bin.
        nbins = _get_axis_len(self, axis) + (1 if is_slice_stop else 0)
        index = index + 1
        if abs(index) > nbins:
            raise IndexError(f"Histogram index {index - 1} out of range for axis {axis}. Valid range: (0,{nbins})")
        return index

    raise index


def _compute_uhi_index(self, index, axis, flow=True):
    """Convert tag functors to valid bin indices."""
    if isinstance(index, _rebin) or index is _sum:
        index = slice(None, None, index)

    if callable(index) or isinstance(index, int):
        return _process_index_for_axis(self, index, axis)

    if isinstance(index, slice):
        start, stop = _resolve_slice_indices(self, index, axis, flow)
        return slice(start, stop, index.step)

    raise TypeError(f"Unsupported index type: {type(index).__name__}")


def _compute_common_index(self, index, flow=True):
    """Normalize and expand the index to match the histogram dimension."""
    dim = self.GetDimension()
    if isinstance(index, dict):
        expanded_index = [slice(None)] * dim
        for axis, value in index.items():
            expanded_index[axis] = value
        index = tuple(expanded_index)

    if not isinstance(index, tuple):
        index = (index,)

    if index.count(...) > 1:
        raise IndexError("Only one ellipsis is allowed in the index.")

    if any(idx is ... for idx in index):
        ellipsis_pos = index.index(...)
        index = index[:ellipsis_pos] + (slice(None),) * (dim - len(index) + 1) + index[ellipsis_pos + 1 :]

    if len(index) != dim:
        raise IndexError(f"Expected {dim} indices, got {len(index)}")

    return [_compute_uhi_index(self, idx, axis, flow) for axis, idx in enumerate(index)]


def _setbin(self, index, value):
    """Set the bin content for a specific bin index"""
    self.SetBinContent(index, value)


def _resolve_slice_indices(self, index, axis, flow=True):
    """Resolve slice start and stop indices for a given axis"""
    start, stop = index.start, index.stop
    start = (
        _process_index_for_axis(self, start, axis) if start is not None else _underflow(self, axis) + (0 if flow else 1)
    )
    stop = (
        _process_index_for_axis(self, stop, axis, is_slice_stop=True)
        if stop is not None
        else _overflow(self, axis) + (1 if flow else 0)
    )
    if start < _underflow(self, axis) or stop > (_overflow(self, axis) + 1) or start > stop:
        raise IndexError(
            f"Slice indices {start, stop} out of range for axis {axis}. Valid range: {_underflow(self, axis), _overflow(self, axis) + 1}"
        )
    return start, stop


def _apply_actions(hist, actions, index, unprocessed_index, original_hist):
    """Apply rebinning or summing actions to the histogram, returns a new histogram"""
    if not actions or all(a is None for a in actions):
        return hist

    if any(a is _sum or a is sum for a in actions):
        sum_axes = tuple(i for i, a in enumerate(actions) if a is _sum or a is sum)
        if original_hist.GetDimension() == 1:
            # For the integral of a 1D histogram, we need special handling for the flow bins
            # h[::sum] is equivalent to h.Integral(0, nbins+1)
            # h[0:len:sum] is equivalent to h.Integral(1, nbins)
            start, stop = index[0].start, index[0].stop
            include_oflow = True if unprocessed_index.stop is None else False
            args = [start, stop - (1 if not include_oflow else 0)]
            hist = _sum(original_hist, sum_axes, args)
        else:
            hist = _sum(hist, sum_axes)

    if any(isinstance(a, _rebin) for a in actions):
        rebins = [a.ngroup if isinstance(a, _rebin) else 1 for a in actions if a is not _sum]
        hist = _rebin(rebins)(hist)

    if any(a is not None and not (isinstance(a, _rebin) or a is _sum or a is sum) for a in actions):
        raise ValueError(f"Unsupported action detected in actions {actions}")

    return hist


def _get_processed_slices(self, index):
    """Process slices and extract actions for each axis"""
    if len(index) != self.GetDimension():
        raise IndexError(f"Expected {self.GetDimension()} indices, got {len(index)}")
    processed_slices, actions = [], [None] * self.GetDimension()
    for axis, idx in enumerate(index):
        if isinstance(idx, slice):
            processed_slices.append((idx.start, idx.stop))
            actions[axis] = idx.step
        elif isinstance(idx, int):
            processed_slices.append((idx, idx + 1))
            actions[axis] = _sum
        else:
            raise TypeError(f"Unsupported index type: {type(idx).__name__}")

    return processed_slices, actions


def _slice_get(self, index, unprocessed_index):
    """
    This method creates a new histogram containing only the data from the
    specified slice.

    Steps:
    - Process the slices and extract the actions for each axis.
    - Get a new sliced histogram.
    - Apply any rebinning or summing actions to the resulting histogram.
    """
    import ROOT

    processed_slices, actions = _get_processed_slices(self, index)
    args_vec = ROOT.std.vector("Int_t")([item for pair in processed_slices for item in pair])

    target_hist = ROOT.Internal.Slice(self, args_vec)

    return _apply_actions(target_hist, actions, index, unprocessed_index, self)


def _slice_set(self, index, unprocessed_index, value):
    """
    This method modifies the histogram by updating the bin contents for the
    specified slice. It supports assigning a scalar value to all bins or
    assigning an array of values, provided the array's shape matches the slice.
    """
    import numpy as np

    import ROOT

    if not np.isscalar(value):
        try:
            value = np.asanyarray(value)
        except AttributeError:
            raise TypeError(f"Unsupported value type: {type(value).__name__}")

    # Depending on the shape of the array provided, we can set or not the flow bins
    # Setting with a scalar does not set the flow bins
    # broadcasting an array to the shape of the slice does not set the flow bins neither
    flow = False
    if isinstance(value, np.ndarray):
        processed_slices, _ = _get_processed_slices(self, index)
        slice_shape = tuple(stop - start for start, stop in processed_slices)
        flow = value.size == np.prod(slice_shape)

    if not flow:
        index = _compute_common_index(self, unprocessed_index, flow=False)

    processed_slices, actions = _get_processed_slices(self, index)
    slice_shape = tuple(stop - start for start, stop in processed_slices)
    slice_edges = ROOT.std.vector("std::pair<Int_t, Int_t>")()
    for start, stop in processed_slices:
        slice_edges.push_back(ROOT.std.pair("Int_t", "Int_t")(start, stop))

    if np.isscalar(value):
        value = ROOT.std.vector("Double_t")([value] * np.prod(slice_shape))
    else:
        if value.size != np.prod(slice_shape):
            try:
                value = np.broadcast_to(value, slice_shape)
            except ValueError:
                raise ValueError(f"Expected {np.prod(slice_shape)} bin values, got {value.size}")
        value = ROOT.std.vector("Double_t")(value.flatten().astype(np.float64))

    ROOT.Internal.SetSliceContent(self, value, slice_edges)

    _apply_actions(self, actions, index, unprocessed_index, self)


def _getitem(self, index):
    uhi_index = _compute_common_index(self, index)
    if all(isinstance(i, int) for i in uhi_index):
        return self.GetBinContent(*uhi_index)
    if any(isinstance(i, slice) for i in uhi_index):
        return _slice_get(self, uhi_index, index)


def _setitem(self, index, value):
    uhi_index = _compute_common_index(self, index)
    if all(isinstance(i, int) for i in uhi_index):
        _setbin(self, self.GetBin(*uhi_index), value)
    elif any(isinstance(i, slice) for i in uhi_index):
        _slice_set(self, uhi_index, index, value)


def _iter(self):
    array = _values_by_copy(self, flow=True)
    for val in array.flat:
        yield val.item()



## \file
## \ingroup tutorial_hist
## \notebook
## Example of function called when a mouse event occurs in a pad.
## When moving the mouse in the canvas, a second canvas shows the
## projection along X of the bin corresponding to the Y position
## of the mouse. The resulting histogram is fitted with a gaussian.
## A "dynamic" line shows the current bin position in Y.
## This more elaborated example can be used as a starting point
## to develop more powerful interactive applications exploiting CINT
## as a development engine.
##
## Note that a class is used to hold on to the canvas that display
## the selected slice.
##
## \macro_image
## \macro_code
##
## \author Rene Brun, Johann Cohen-Tanugi, Wim Lavrijsen, Enric Tejedor

import sys
import ctypes

from ROOT import gRandom, gPad, gROOT, gVirtualX
from ROOT import kTRUE, kRed
from ROOT import TCanvas, TH2, TH2F


class DynamicExec:

   def __init__( self ):
      self._cX   = None
      self._cY   = None
      self._old  = None

   def __call__( self ):

      h = gPad.GetSelected();
      if not h:
         return

      if not isinstance( h, TH2 ):
         return

      gPad.GetCanvas().FeedbackMode( kTRUE )

    # erase old position and draw a line at current position
      px = gPad.GetEventX()
      py = gPad.GetEventY()

      uxmin, uxmax = gPad.GetUxmin(), gPad.GetUxmax()
      uymin, uymax = gPad.GetUymin(), gPad.GetUymax()
      pxmin, pxmax = gPad.XtoAbsPixel( uxmin ), gPad.XtoAbsPixel( uxmax )
      pymin, pymax = gPad.YtoAbsPixel( uymin ), gPad.YtoAbsPixel( uymax )

      if self._old != None:
         gVirtualX.DrawLine( pxmin, self._old[1], pxmax, self._old[1] )
         gVirtualX.DrawLine( self._old[0], pymin, self._old[0], pymax )
      gVirtualX.DrawLine( pxmin, py, pxmax, py )
      gVirtualX.DrawLine( px, pymin, px, pymax )

      self._old = px, py

      upx = gPad.AbsPixeltoX( px )
      x = gPad.PadtoX( upx )
      upy = gPad.AbsPixeltoY( py )
      y = gPad.PadtoY( upy )

      padsav = gPad

    # create or set the display canvases
      if not self._cX:
         self._cX = TCanvas( 'c2', 'Projection Canvas in X', 730, 10, 700, 500 )
      else:
         self._DestroyPrimitive( 'X' )

      if not self._cY:
         self._cY = TCanvas( 'c3', 'Projection Canvas in Y', 10, 550, 700, 500 )
      else:
         self._DestroyPrimitive( 'Y' )

      self.DrawSlice( h, y, 'Y' )
      self.DrawSlice( h, x, 'X' )

      padsav.cd()

   def _DestroyPrimitive( self, xy ):
      proj = getattr( self, '_c'+xy ).GetPrimitive( 'Projection '+xy )
      if proj:
         proj.IsA().Destructor( proj )

   def DrawSlice( self, histo, value, xy ):
      yx = xy == 'X' and 'Y' or 'X'

    # draw slice corresponding to mouse position
      canvas = getattr( self, '_c'+xy )
      canvas.SetGrid()
      canvas.cd()

      bin = getattr( histo, 'Get%saxis' % xy )().FindBin( value )
      hp = getattr( histo, 'Projection' + yx )( '', bin, bin )
      hp.SetFillColor( 38 )
      hp.SetName( 'Projection ' + xy )
      hp.SetTitle( xy + 'Projection of bin=%d' % bin )
      hp.Fit( 'gaus', 'ql' )
      hp.GetFunction( 'gaus' ).SetLineColor( kRed )
      hp.GetFunction( 'gaus' ).SetLineWidth( 6 )
      canvas.Update()


if __name__ == '__main__':
 # create a new canvas.
   c1 = TCanvas('c1', 'Dynamic Slice Example', 10, 10, 700, 500 )
   c1.SetFillColor( 42 )
   c1.SetFrameFillColor( 33 )

 # create a 2-d histogram, fill and draw it
   hpxpy  = TH2F( 'hpxpy', 'py vs px', 40, -4, 4, 40, -4, 4 )
   hpxpy.SetStats( 0 )
   x, y = ctypes.c_double( 0.1 ), ctypes.c_double( 0.101 )
   for i in range( 50000 ):
   # pass ctypes doubles by reference, then retrieve their modified values with .value
     gRandom.Rannor( x, y )
     hpxpy.Fill( x.value, y.value )
   hpxpy.Draw( 'COL' )

 # Add a TExec object to the canvas (explicit use of __main__ is for IPython)
   import __main__
   __main__.slicer = DynamicExec()
   c1.AddExec( 'dynamic', 'TPython::Exec( "slicer()" );' )
   c1.Update()



## \file
## \ingroup tutorial_dataframe
## \notebook -draw
## A single top analysis using the ATLAS Open Data release of 2020, with RDataFrame.
##
## This tutorial is the analysis of single top production adapted from the ATLAS Open Data release in 2020
## (http://opendata.atlas.cern/release/2020/documentation/). The data was recorded with the ATLAS detector
## during 2016 at a center-of-mass energy of 13 TeV. Top quarks with a mass of about 172 GeV are mostly
## produced in pairs but also appear alone, dominantly from the decays of a W boson in association with a light jet.
##
## The analysis is translated to a RDataFrame workflow processing up to 60 GB of simulated events and data.
## By default the analysis runs on a preskimmed dataset to reduce the runtime. The full dataset can be used with
## the --full-dataset argument and you can also run only on a fraction of the original dataset using the argument --lumi-scale.
##
## See the [corresponding spec json file](https://github.com/root-project/root/blob/master/tutorials/analysis/dataframe/df107_SingleTopAnalysis.json).
##
## \macro_image
## \macro_code
## \macro_output
##
## \date July 2020
## \author Stefan Wunsch (KIT, CERN)

import ROOT
import sys
import json
import argparse
import os

# Argument parsing
parser = argparse.ArgumentParser()
parser.add_argument("--lumi-scale", type=float, default=0.05,
                    help="Run only on a fraction of the total available 10 fb^-1 (only usable together with --full-dataset)")
parser.add_argument("--full-dataset", action="store_true", default=False,
                    help="Use the full dataset (use --lumi-scale to run only on a fraction of it)")
parser.add_argument("-b", action="store_true", default=False, help="Use ROOT batch mode")
parser.add_argument("-t", action="store_true", default=False, help="Use implicit multi threading (for the full dataset only possible with --lumi-scale 1.0)")
if 'df107_SingleTopAnalysis.py' in sys.argv[0]:
    # Script
    args = parser.parse_args()
else:
    # Notebook
    args = parser.parse_args(args=[])

if args.b: ROOT.gROOT.SetBatch(True)
if args.t: ROOT.EnableImplicitMT()

if not args.full_dataset: lumi_scale = 0.05 # The preskimmed dataset contains only 0.5 fb^-1
else: lumi_scale = args.lumi_scale
lumi = 10064.0
print('Run on data corresponding to {:.1f} fb^-1 ...'.format(lumi * lumi_scale / 1000.0))

if args.full_dataset: dataset_path = "root://eospublic.cern.ch//eos/opendata/atlas/OutreachDatasets/2020-01-22"
else: dataset_path = "root://eospublic.cern.ch//eos/root-eos/reduced_atlas_opendata/singletop"

# Create a ROOT dataframe for each dataset
# Note that we load the filenames from the external json file placed in the same folder than this script.
files = json.load(open(os.path.join(ROOT.gROOT.GetTutorialsDir(), "analysis/dataframe/df107_SingleTopAnalysis.json")))
processes = files.keys()
df = {}
xsecs = {}
sumws = {}
samples = []
for p in processes:
    for d in files[p]:
        # Construct the dataframes
        folder = d[0] # Folder name
        sample = d[1] # Sample name
        xsecs[sample] = d[2] # Cross-section
        sumws[sample] = d[3] # Sum of weights
        num_events = d[4] # Number of events
        samples.append(sample)
        df[sample] = ROOT.RDataFrame("mini", "{}/1lep/{}/{}.1lep.root".format(dataset_path, folder, sample))

        # Scale down the datasets if requested
        if args.full_dataset and lumi_scale < 1.0:
            df[sample] = df[sample].Range(int(num_events * lumi_scale))

# Select events for the analysis and make histograms of the top mass

# Just-in-time compile custom helper function performing complex computations
ROOT.gInterpreter.Declare("""
using cRVecF = const ROOT::RVecF &;
using cRVecI = const ROOT::RVecI &;
int FindGoodLepton(cRVecI goodlep, cRVecI type, cRVecF lep_pt, cRVecF lep_eta, cRVecF lep_phi, cRVecF lep_e, cRVecF trackd0pv, cRVecF tracksigd0pv, cRVecF z0)
{
    int idx = -1; // Return -1 if no good lepton is found.
    for(auto i = 0; i < type.size(); i++) {
        if(!goodlep[i]) continue;
        if (type[i] == 11 && abs(lep_eta[i]) < 2.47 && (abs(lep_eta[i]) < 1.37 || abs(lep_eta[i]) > 1.52) && abs(trackd0pv[i] / tracksigd0pv[i]) < 5) {
            const ROOT::Math::PtEtaPhiEVector p(lep_pt[i], lep_eta[i], lep_phi[i], lep_e[i]);
            if (abs(z0[i] * sin(p.Theta())) < 0.5) {
                if (idx == -1) idx = i;
                else return -1; // Accept only events with exactly one good lepton
            }
        }
        if (type[i] == 13 && abs(lep_eta[i]) < 2.5 && abs(trackd0pv[i] / tracksigd0pv[i]) < 3) {
            const ROOT::Math::PtEtaPhiEVector p(lep_pt[i], lep_eta[i], lep_phi[i], lep_e[i]);
            if (abs(z0[i] * sin(p.Theta())) < 0.5) {
                if (idx == -1) idx = i;
                else return -1; // Accept only events with exactly one good lepton
            }
        }
    }
    return idx;
}
""")

for s in samples:
    # Select events with electron or muon trigger and with a missing transverse energy above 30 GeV
    df[s] = df[s].Filter("trigE || trigM")\
                 .Filter("met_et > 30000")

    # Perform preselection of highly isolated leptons
    df[s] = df[s].Define("goodlep", "lep_isTightID && lep_pt > 35000 && lep_ptcone30 / lep_pt < 0.1 && lep_etcone20 / lep_pt < 0.1")\
                 .Filter("ROOT::VecOps::Sum(goodlep) > 0")

    # Find a single good lepton, otherwise return -1 as index
    df[s] = df[s].Define("idx_lep", "FindGoodLepton(goodlep, lep_type, lep_pt, lep_eta, lep_phi, lep_E, lep_trackd0pvunbiased, lep_tracksigd0pvunbiased, lep_z0)")\
                 .Filter("idx_lep != -1")

    # Compute transverse mass of the W boson using the missing transverse energy and the good lepton
    # Use only events with a transverse mass of the reconstructed W boson larger than 60 GeV
    df[s] = df[s].Define("mtw", "sqrt(2 * lep_pt[idx_lep] * met_et * (1 - cos(lep_phi[idx_lep] - met_phi)))")\
                 .Filter("mtw > 60000")

    # Perform preselection of jets
    df[s] = df[s].Filter("ROOT::VecOps::Sum(jet_pt > 30000 && abs(jet_eta) < 2.5) > 0")

    # Select events with two good jets and one b-jet and find the indices in the collections
    df[s] = df[s].Define("goodjet", "jet_pt > 60000 || abs(jet_eta) > 2.4 || jet_jvt > 0.59")\
                 .Filter("ROOT::VecOps::Sum(goodjet) == 2")\
                 .Define("goodbjet", "goodjet && jet_MV2c10 > 0.8244273")\
                 .Filter("ROOT::VecOps::Sum(goodbjet) == 1")\
                 .Define("idx_tagged", "ROOT::VecOps::ArgMax(goodjet && goodbjet)")\
                 .Define("idx_untagged", "ROOT::VecOps::ArgMax(goodjet && !goodbjet)")

    # Select events based on the jet kinematics and the scalar sum of the transverse momentum
    # from the lepton, jets and met above 195 GeV
    df[s] = df[s].Filter("abs(jet_eta[idx_untagged]) > 1.5 && abs(jet_eta[idx_tagged] - jet_eta[idx_untagged]) > 1.5")\
                 .Filter("lep_pt[idx_lep] + jet_pt[idx_tagged] + jet_pt[idx_untagged] + met_et > 195000")

# Compute luminosity, scale factors and MC weights for simulated events
for s in samples:
    if "data" in s:
        df[s] = df[s].Define("weight", "1.0")
    else:
        # The single top MC weights are either 1 or -1
        if "single" in s: stop_norm = "mcWeight / abs(mcWeight)"
        else: stop_norm = "mcWeight"
        df[s] = df[s].Define("weight", "scaleFactor_ELE * scaleFactor_MUON * scaleFactor_LepTRIGGER * scaleFactor_PILEUP * scaleFactor_BTAG * {} * {} / {} * {}".format(stop_norm, xsecs[s], sumws[s], lumi))

# Reconstruct the top mass from the lepton, the missing transverse energy and the b-jet

# Just-in-time compile the function to compute the top mass from the constituents
ROOT.gInterpreter.Declare("""
float ComputeTopMass(float lep_pt, float lep_eta, float lep_phi, float lep_e, float jet_pt, float jet_eta, float jet_phi, float jet_e, float met_et, float met_phi)
{
    const ROOT::Math::PtEtaPhiEVector lep(lep_pt / 1000.0, lep_eta, lep_phi, lep_e / 1000.0);
    const ROOT::Math::PtEtaPhiEVector met(met_et / 1000.0, 0, met_phi, met_et / 1000.0);
    const ROOT::Math::PtEtaPhiEVector bjet(jet_pt / 1000.0, jet_eta, jet_phi, jet_e / 1000.0);
    // Please note that we treat here the missing transverse energy as the neutrino, even though the z component is missing!
    return (lep + met + bjet).M();
}
""")

histos = {}
for s in samples:
    df[s] = df[s].Define("top_mass", "ComputeTopMass(lep_pt[idx_lep], lep_eta[idx_lep], lep_phi[idx_lep], lep_E[idx_lep], jet_pt[idx_tagged], jet_eta[idx_tagged], jet_phi[idx_tagged], jet_E[idx_tagged], met_et, met_phi)")
    histos[s] = df[s].Histo1D(ROOT.RDF.TH1DModel("top_mass", "", 10, 100, 400), "top_mass", "weight")

# Run the event loop and merge histograms of the respective processes

# RunGraphs allows to run the event loops of the separate RDataFrame graphs
# concurrently. This results in an improved usage of the available resources
# if each separate RDataFrame can not utilize all available resources, e.g.,
# because not enough data is available.
ROOT.RDF.RunGraphs([histos[s] for s in samples])

def merge_histos(label):
    h = None
    for i, d in enumerate(files[label]):
        t = histos[d[1]].GetValue()
        if i == 0: h = t.Clone()
        else: h.Add(t)
    h.SetNameTitle(label, label)
    return h

data = merge_histos("data")
twtb = merge_histos("twtb")
singletop = merge_histos("singletop")
wjets = merge_histos("wjets")

# Create the plot

# Set styles
ROOT.gROOT.SetStyle("ATLAS")

# Create canvas with pad
c = ROOT.TCanvas("c", "", 600, 600)
pad = ROOT.TPad("upper_pad", "", 0, 0, 1, 1)
pad.SetTickx(False)
pad.SetTicky(False)
pad.Draw()
pad.cd()

# Draw stack with MC contributions
stack = ROOT.THStack()
wjets.Scale(1.1) # Corrected normalization derived from a validation region
for h, color in zip(
        [wjets, twtb, singletop],
        [(0.87, 0.35, 0.42), (0.61, 0.6, 0.8), (0.82, 0.94, 0.76)]):
    h.SetLineWidth(1)
    h.SetLineColor("black")
    h.SetFillColor(ROOT.TColor.GetColor(*color))
    # From Python 3.11 onward, implicit conversion of the tuple works too:
    # h.SetFillColor(color)
    stack.Add(h)
stack.Draw("HIST")
stack.GetXaxis().SetTitle("m_{W(l#nu)+b} [GeV]")
stack.GetYaxis().SetTitle("Events")
stack.GetYaxis().SetLabelSize(0.04)
stack.GetYaxis().SetTitleSize(0.045)
stack.GetXaxis().SetLabelSize(0.04)
stack.GetXaxis().SetTitleSize(0.045)
stack.SetMinimum(0)
stack.SetMaximum(5000 * lumi_scale)
stack.GetYaxis().ChangeLabel(1, -1, 0)

# Draw data
data.SetMarkerStyle(20)
data.SetMarkerSize(1.2)
data.SetLineWidth(2)
data.SetLineColor("black")
data.Draw("E SAME")

# Add legend
legend = ROOT.TLegend(0.60, 0.65, 0.92, 0.92)
legend.SetTextFont(42)
legend.SetFillStyle(0)
legend.SetBorderSize(0)
legend.SetTextSize(0.035)
legend.SetTextAlign(32)
legend.AddEntry(data, "Data" ,"lep")
legend.AddEntry(singletop, "Single top + jet", "f")
legend.AddEntry(twtb, "t#bar{t},Wt,t#bar{b}", "f")
legend.AddEntry(wjets, "W+jets", "f")
legend.Draw("SAME")

# Add ATLAS label
text = ROOT.TLatex()
text.SetNDC()
text.SetTextFont(72)
text.SetTextSize(0.045)
text.DrawLatex(0.21, 0.86, "ATLAS")
text.SetTextFont(42)
text.DrawLatex(0.21 + 0.16, 0.86, "Open Data")
text.SetTextSize(0.04)
text.DrawLatex(0.21, 0.80, "#sqrt{{s}} = 13 TeV, {:.1f} fb^{{-1}}".format(lumi * lumi_scale / 1000.0))

# Save the plot
c.SaveAs("df107_SingleTopAnalysis.png")
print("Saved figure to df107_SingleTopAnalysis.png")



# Author: Sergey Linev GSI  11/2025

################################################################################
# Copyright (C) 1995-2025, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

r'''
\pythondoc TBrowser

Functionality of constructor and Draw() methods were extended to support interactive
work in the python scripts. If extra block parameter is True, script execution
will be suspended until <space> key pressed by user. Simple example:

\code{.py}
import ROOT

# block until space is pressed
br = ROOT.TBrowser(block = True)

# continue work to load new files

# block here again until space is pressed
br.Draw(block = True)

# continues after <space> is pressed
\endcode
\endpythondoc
'''

from . import _run_root_event_loop, pythonization


def _TBrowser_constructor(self, *args, block: bool = False):

   self._original_constructor(*args)

   if block:
      _run_root_event_loop()


def _TBrowser_Draw(self, option: str = "", block: bool = False):
   """
   Draw the browser.
   Also blocks script execution and runs the ROOT graphics event loop until the <space> keyword is pressed,
   but only if the following conditions are met:
   * The `block` optional argument is set to `True`.
   * ROOT graphics are enabled, i.e. `ROOT.gROOT.IsBatch() == False`.
   * The script is running not in ipython notebooks.
   """

   self._Draw(option)

   # run loop if block flag is set
   if block:
      _run_root_event_loop()


@pythonization('TBrowser')
def pythonize_tbrowser(klass):
    # Parameters:
    # klass: class to be pythonized

    klass._original_constructor = klass.__init__
    klass.__init__ = _TBrowser_constructor

    klass._Draw = klass.Draw
    klass.Draw = _TBrowser_Draw

